from abc import ABC, abstractmethod

import aiohttp


class HttpClient(ABC):
    @abstractmethod
    def fetch(self, url: str, options: dict) -> "HttpResponse":
        pass


class HttpResponse:
    def __init__(self, ok: bool, status_code: int, json_data: dict):
        self.ok = ok
        self.status_code = status_code
        self._json_data = json_data

    def json(self):
        return self._json_data


class DefaultHttpClient(HttpClient):
    async def fetch(self, url: str, options: dict) -> HttpResponse:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=options["method"],
                url=url,
                headers=options.get("headers", {}),
                json=options.get("data", None),
            ) as response:
                try:
                    json_data = await response.json()
                    return HttpResponse(
                        ok=response.status >= 200 and response.status <= 299,
                        status_code=response.status,
                        json_data={
                            'data': json_data
                        },
                    )
                except Exception as e:
                    return HttpResponse(
                        ok=False,
                        status_code=response.status,
                        json_data={},
                    )


def default_http_client() -> HttpClient:
    return DefaultHttpClient()



import pytest

from bsv.signed_message import SignedMessage
from bsv.keys import PrivateKey


def test_signs_message_for_recipient():
    sender = PrivateKey(15)
    recipient = PrivateKey(21)
    recipient_pub = recipient.public_key()
    message = bytes([1, 2, 4, 8, 16, 32])
    signature = SignedMessage.sign(message, sender, verifier=recipient_pub)
    verified = SignedMessage.verify(message, signature, recipient=recipient)
    assert verified is True

def test_signs_message_for_anyone():
    sender = PrivateKey(15)
    message = bytes([1, 2, 4, 8, 16, 32])
    signature = SignedMessage.sign(message, sender)
    verified = SignedMessage.verify(message, signature)
    assert verified is True

def test_fails_to_verify_message_with_wrong_version():
    sender = PrivateKey(15)
    recipient = PrivateKey(21)
    recipient_pub = recipient.public_key()
    message = bytes([1, 2, 4, 8, 16, 32])
    signature = bytearray(SignedMessage.sign(message, sender, verifier=recipient_pub))
    signature[0] = 1  # Altering the version byte
    with pytest.raises(ValueError, match=r'Message version mismatch: Expected 42423301, received 01423301'):
        SignedMessage.verify(message, signature, recipient=recipient)

def test_fails_to_verify_message_with_no_verifier_when_required():
    sender = PrivateKey(15)
    recipient = PrivateKey(21)
    recipient_pub = recipient.public_key()
    message = bytes([1, 2, 4, 8, 16, 32])
    signature = SignedMessage.sign(message, sender, verifier=recipient_pub)
    with pytest.raises(ValueError, match=r'This signature can only be verified with knowledge of a specific private key\. The associated public key is: .*'):
        SignedMessage.verify(message, signature)

def test_fails_to_verify_message_with_wrong_verifier():
    sender = PrivateKey(15)
    recipient = PrivateKey(21)
    wrong_recipient = PrivateKey(22)
    recipient_pub = recipient.public_key()
    message = bytes([1, 2, 4, 8, 16, 32])
    signature = SignedMessage.sign(message, sender, verifier=recipient_pub)
    with pytest.raises(ValueError, match=r'The recipient public key is .* but the signature requires the recipient to have public key .*'):
        SignedMessage.verify(message, signature, recipient=wrong_recipient)



from abc import ABC, abstractmethod
from typing import Awaitable

class FeeModel(ABC):
    """
    Represents the interface for a transaction fee model.
    This interface defines a standard method for computing a fee when given a transaction.

    @interface
    @property {function} computeFee - A function that takes a Transaction object and returns an integer representing the number of satoshis the transaction should cost.
    """
    
    @abstractmethod
    def compute_fee(self, transaction) -> int:
        pass


import hmac
from hashlib import sha512
from typing import Union

from ..base58 import base58check_decode, base58check_encode
from ..constants import BIP32_SEED_BYTE_LENGTH
from ..constants import NETWORK_XPUB_PREFIX_DICT, NETWORK_XPRV_PREFIX_DICT
from ..constants import Network, XKEY_BYTE_LENGTH, XKEY_PREFIX_LIST, PUBLIC_KEY_COMPRESSED_PREFIX_LIST
from ..constants import XPUB_PREFIX_NETWORK_DICT, XPRV_PREFIX_NETWORK_DICT
from ..curve import curve, curve_add, curve_multiply
from ..keys import PublicKey, PrivateKey


class Xkey:
    """
    [  : 4] prefix
    [ 4: 5] depth
    [ 5: 9] parent public key fingerprint
    [ 9:13] child index
    [13:45] chain code
    [45:78] key (private/public)
    """

    def __init__(self, xkey: Union[str, bytes]):
        if isinstance(xkey, str):
            self.payload: bytes = base58check_decode(xkey)
        elif isinstance(xkey, bytes):
            self.payload: bytes = xkey
        else:
            raise TypeError('unsupported extended key type')

        assert len(self.payload) == XKEY_BYTE_LENGTH, 'invalid extended key length'
        self.prefix: bytes = self.payload[:4]
        self.depth: int = self.payload[4]
        self.fingerprint: bytes = self.payload[5:9]
        self.index: int = int.from_bytes(self.payload[9:13], 'big')
        self.chain_code: bytes = self.payload[13:45]
        self.key_bytes: bytes = self.payload[45:]
        assert self.prefix in XKEY_PREFIX_LIST, 'invalid extended key prefix'

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Xkey):
            return self.payload == o.payload
        return super().__eq__(o)  # pragma: no cover

    def __str__(self) -> str:
        return base58check_encode(self.payload)

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()


class Xpub(Xkey):

    def __init__(self, xpub: Union[str, bytes]):
        super().__init__(xpub)
        self.network: Network = XPUB_PREFIX_NETWORK_DICT.get(self.prefix)
        assert self.network, 'unknown xpub prefix'
        assert self.payload[45:46] in PUBLIC_KEY_COMPRESSED_PREFIX_LIST, 'invalid public key in xpub'
        self.key: PublicKey = PublicKey(self.key_bytes)

    def ckd(self, index: Union[int, str, bytes]) -> 'Xpub':
        if isinstance(index, int):
            index = index.to_bytes(4, 'big')
        elif isinstance(index, str):
            index = bytes.fromhex(index)
        assert len(index) == 4, 'index should be a 4 bytes integer'
        assert index[0] < 0x80, "can't make hardened derivation from xpub"

        payload: bytes = self.prefix
        payload += (self.depth + 1).to_bytes(1, 'big')
        payload += self.key.hash160()[:4]
        payload += index

        h: bytes = hmac.new(self.chain_code, self.key.serialize() + index, sha512).digest()
        offset: int = int.from_bytes(h[:32], 'big')
        child: PublicKey = PublicKey(curve_add(self.key.point(), curve_multiply(offset, curve.g)))

        payload += h[32:]
        payload += child.serialize()

        return Xpub(payload)

    def public_key(self) -> PublicKey:
        return self.key

    def address(self) -> str:
        return self.key.address(network=self.network)

    @classmethod
    def from_xprv(cls, xprv: Union[str, bytes, 'Xprv']) -> 'Xpub':
        if not isinstance(xprv, Xprv):
            xprv = Xprv(xprv)
        payload: bytes = NETWORK_XPUB_PREFIX_DICT.get(xprv.network)
        payload += xprv.depth.to_bytes(1, 'big')
        payload += xprv.fingerprint
        payload += xprv.index.to_bytes(4, 'big')
        payload += xprv.chain_code
        payload += xprv.key.public_key().serialize()
        return Xpub(payload)


class Xprv(Xkey):

    def __init__(self, xprv: Union[str, bytes]):
        super().__init__(xprv)
        self.network: Network = XPRV_PREFIX_NETWORK_DICT.get(self.prefix)
        assert self.network, 'unknown xprv prefix'
        assert self.payload[45] == 0, 'invalid private key in xprv'
        self.key: PrivateKey = PrivateKey(self.key_bytes[1:], network=self.network)

    def ckd(self, index: Union[int, str, bytes]) -> 'Xprv':
        if isinstance(index, int):
            index = index.to_bytes(4, 'big')
        elif isinstance(index, str):
            index = bytes.fromhex(index)
        assert len(index) == 4, 'index should be a 4 bytes integer'

        payload: bytes = self.prefix
        payload += (self.depth + 1).to_bytes(1, 'big')
        payload += self.key.public_key().hash160()[:4]
        payload += index

        message: bytes = (self.key.public_key().serialize() if index[0] < 0x80 else self.key_bytes) + index
        h: bytes = hmac.new(self.chain_code, message, sha512).digest()
        offset: int = int.from_bytes(h[:32], 'big')
        child: PrivateKey = PrivateKey((self.key.int() + offset) % curve.n)

        payload += h[32:]
        payload += b'\x00' + child.serialize()

        return Xprv(payload)

    def xpub(self) -> Xpub:
        return Xpub.from_xprv(self)

    def private_key(self) -> PrivateKey:
        return self.key

    def public_key(self) -> PublicKey:
        return self.key.public_key()

    def address(self) -> str:
        return self.key.address()

    @classmethod
    def from_seed(cls, seed: Union[str, bytes], network: Network = Network.MAINNET):
        """
        derive master extended private key from seed
        """
        if isinstance(seed, str):
            seed = bytes.fromhex(seed)
        assert len(seed) == BIP32_SEED_BYTE_LENGTH, 'invalid seed byte length'

        payload: bytes = NETWORK_XPRV_PREFIX_DICT.get(network)
        payload += b'\x00'
        payload += b'\x00\x00\x00\x00'
        payload += b'\x00\x00\x00\x00'

        h: bytes = hmac.new(b'Bitcoin seed', seed, sha512).digest()
        payload += h[32:]
        payload += b'\x00' + h[:32]

        return Xprv(payload)


def step_to_index(step: Union[str, int]) -> int:
    """
    convert step (sub path) normal derivation or hardened derivation into child index
    """
    assert type(step).__name__ in ['str', 'int'], 'unsupported step type'
    if isinstance(step, str):
        assert len(step), 'invalid step'
        hardened: bool = step[-1] == "'"
        index: int = (0x80000000 if hardened else 0) + int(step[:-1] if hardened else step)
    else:
        index: int = step
    assert 0 <= index < 0xffffffff, 'step out of range'
    return index


def ckd(xkey: Union[Xprv, Xpub], path: str) -> Union[Xprv, Xpub]:
    """
    derive an extended key according to path like "m/44'/0'/1'/0/10" (absolute) or "./0/10" (relative)
    """
    steps = path.strip(' ').strip('/').split('/')
    assert steps and steps[0] in ['m', '.']

    if steps[0] == 'm':
        # should be master key
        assert (
                xkey.depth == 0 and xkey.fingerprint == b'\x00\x00\x00\x00' and xkey.index == 0
        ), 'absolute path for non-master key'

    child = xkey
    for step in steps[1:]:
        child = child.ckd(step_to_index(step))
    return child


def master_xprv_from_seed(seed: Union[str, bytes], network: Network = Network.MAINNET) -> Xprv:
    return Xprv.from_seed(seed, network)



from secrets import randbits

import pytest

from bsv.aes_cbc import InvalidPadding
from bsv.aes_cbc import append_pkcs7_padding, strip_pkcs7_padding, aes_encrypt_with_iv, aes_decrypt_with_iv


def test():
    message: bytes = b'hello world'
    padding_message: bytes = b'hello world\x05\x05\x05\x05\x05'
    assert append_pkcs7_padding(message) == padding_message
    assert strip_pkcs7_padding(padding_message) == message

    message: bytes = b'\x00' * 16
    padding_message: bytes = message + b'\x10' * 16
    assert append_pkcs7_padding(message) == padding_message
    assert strip_pkcs7_padding(padding_message) == message

    with pytest.raises(InvalidPadding, match=r'invalid length'):
        strip_pkcs7_padding(b'')
    with pytest.raises(InvalidPadding, match=r'invalid length'):
        strip_pkcs7_padding(b'\x00' * 15)
    with pytest.raises(InvalidPadding, match=r'invalid padding byte \(out of range\)'):
        strip_pkcs7_padding(b'hello world\x05\x05\x05\x05\xff')
    with pytest.raises(InvalidPadding, match=r'invalid padding byte \(inconsistent\)'):
        strip_pkcs7_padding(b'hello world\x05\x05\x05\x04\x05')

    key_byte_length = 16
    key = randbits(key_byte_length * 8).to_bytes(key_byte_length, 'big')
    iv = randbits(key_byte_length * 8).to_bytes(key_byte_length, 'big')
    encrypted: bytes = aes_encrypt_with_iv(key, iv, message)
    assert message == aes_decrypt_with_iv(key, iv, encrypted)



from contextlib import suppress
from typing import Literal

from .script import ScriptChunk, Script
from ..constants import OpCode, OPCODE_VALUE_NAME_DICT, SIGHASH
from ..curve import curve
from ..hash import sha1, sha256, ripemd160, hash256, hash160
from ..keys import PublicKey
from ..transaction_input import TransactionInput
from ..transaction_preimage import tx_preimage
from ..utils import unsigned_to_bytes, deserialize_ecdsa_der

MAX_SCRIPT_ELEMENT_SIZE = 1024 * 1024 * 1024
MAX_MULTISIG_KEY_COUNT = pow(2, 31) - 1
REQUIRE_MINIMAL_PUSH = True
REQUIRE_PUSH_ONLY_UNLOCKING_SCRIPTS = True
REQUIRE_LOW_S_SIGNATURES = True
REQUIRE_CLEAN_STACK = True


class Spend:
    def __init__(self, params):
        """
        Constructs a Spend object with necessary transaction details.

        :param str params['sourceTXID']: The transaction ID of the source UTXO.
        :param int params['sourceOutputIndex']: The index of the output in the source transaction.
        :param BigNumber params['sourceSatoshis']: The amount of satoshis in the source UTXO.
        :param LockingScript params['lockingScript']: The locking script associated with the UTXO.
        :param int params['transactionVersion']: The version of the current transaction.
        :param list params['otherInputs']: An array of other inputs in the transaction.
        :param list params['outputs']: The outputs of the current transaction.
        :param int params['inputIndex']: The index of this input in the current transaction.
        :param UnlockingScript params['unlockingScript']: The unlocking script for this spend.
        :param int params['inputSequence']: The sequence number of this input.
        :param int params['lockTime']: The lock time of the transaction.

        Example:
        spend = Spend({
            'sourceTXID': "abcd1234",  # sourceTXID
            'sourceOutputIndex': 0,  # sourceOutputIndex
            'sourceSatoshis': BigNumber(1000),  # sourceSatoshis
            'lockingScript': LockingScript.from_asm("OP_DUP OP_HASH160 abcd1234... OP_EQUALVERIFY OP_CHECKSIG"),
            'transactionVersion': 1,  # transactionVersion
            'otherInputs': [{'sourceTXID': "abcd1234", 'sourceOutputIndex': 1, 'sequence': 0xffffffff}],  # otherInputs
            'outputs': [{'satoshis': BigNumber(500), 'lockingScript': LockingScript.from_asm("OP_DUP...")}],  # outputs
            'inputIndex': 0,  # inputIndex
            'unlockingScript': UnlockingScript.from_asm("3045... 02ab..."),
            'inputSequence': 0xffffffff,  # inputSequence
            'lockTime': 0  # lockTime
        })
        """
        self.source_txid = params['sourceTXID']
        self.source_output_index = params['sourceOutputIndex']
        self.source_satoshis = params['sourceSatoshis']
        self.locking_script: Script = params['lockingScript']
        self.transaction_version = params['transactionVersion']
        self.other_inputs = params['otherInputs']
        self.outputs = params['outputs']
        self.input_index = params['inputIndex']
        self.unlocking_script: Script = params['unlockingScript']
        self.input_sequence = params['inputSequence']
        self.lock_time = params['lockTime']

        self.context: Literal["UnlockingScript", "LockingScript"] = 'UnlockingScript'
        self.program_counter = 0
        self.last_code_separator = None
        self.stack = []
        self.alt_stack = []
        self.if_stack = []

    def step(self) -> None:
        # If the context is UnlockingScript, and we have reached the end,
        # set the context to LockingScript and zero the program counter
        if self.context == 'UnlockingScript' and self.program_counter >= len(self.unlocking_script.chunks):
            self.context = 'LockingScript'
            self.program_counter = 0

        if self.context == 'UnlockingScript':
            operation = self.unlocking_script.chunks[self.program_counter]
        else:
            operation = self.locking_script.chunks[self.program_counter]

        is_script_executing = not (b'' in self.if_stack)

        # Read instruction
        current_opcode = operation.op
        if current_opcode not in OPCODE_VALUE_NAME_DICT.keys() \
                and not (b'\x01' <= current_opcode < OpCode.OP_PUSHDATA1):
            self.script_evaluation_error(f'An opcode is missing in this chunk of the {self.context}!')
        if operation.data is not None and len(operation.data) > MAX_SCRIPT_ELEMENT_SIZE:
            _m = f"It's not currently possible to push data larger than {MAX_SCRIPT_ELEMENT_SIZE} bytes."
            self.script_evaluation_error(_m)
        if is_script_executing and self.is_opcode_disabled(current_opcode):
            self.script_evaluation_error('This opcode is currently disabled.')

        if is_script_executing and OpCode.OP_0 <= current_opcode <= OpCode.OP_PUSHDATA4:
            if REQUIRE_MINIMAL_PUSH and not self.is_chunk_minimal(operation):
                self.script_evaluation_error('This data is not minimally-encoded.')
            if operation.data is None:
                self.stack.append(b'')
            else:
                self.stack.append(operation.data)
        elif is_script_executing or (OpCode.OP_IF <= current_opcode <= OpCode.OP_ENDIF):
            if current_opcode in [
                OpCode.OP_1NEGATE,
                OpCode.OP_1,
                OpCode.OP_2,
                OpCode.OP_3,
                OpCode.OP_4,
                OpCode.OP_5,
                OpCode.OP_6,
                OpCode.OP_7,
                OpCode.OP_8,
                OpCode.OP_9,
                OpCode.OP_10,
                OpCode.OP_11,
                OpCode.OP_12,
                OpCode.OP_13,
                OpCode.OP_14,
                OpCode.OP_15,
                OpCode.OP_16,
            ]:
                n = int.from_bytes(current_opcode, 'big') - (int.from_bytes(OpCode.OP_1, 'big') - 1)
                self.stack.append(self.minimally_encode(n))

            elif current_opcode in [
                OpCode.OP_NOP,
                OpCode.OP_NOP1,
                OpCode.OP_NOP2,
                OpCode.OP_NOP3,
                OpCode.OP_NOP4,
                OpCode.OP_NOP5,
                OpCode.OP_NOP6,
                OpCode.OP_NOP7,
                OpCode.OP_NOP8,
                OpCode.OP_NOP9,
                OpCode.OP_NOP10,
                OpCode.OP_NOP11,
                OpCode.OP_NOP12,
                OpCode.OP_NOP13,
                OpCode.OP_NOP14,
                OpCode.OP_NOP15,
                OpCode.OP_NOP16,
                OpCode.OP_NOP17,
                OpCode.OP_NOP18,
                OpCode.OP_NOP19,
                OpCode.OP_NOP20,
                OpCode.OP_NOP21,
                OpCode.OP_NOP22,
                OpCode.OP_NOP23,
                OpCode.OP_NOP24,
                OpCode.OP_NOP25,
                OpCode.OP_NOP26,
                OpCode.OP_NOP27,
                OpCode.OP_NOP28,
                OpCode.OP_NOP29,
                OpCode.OP_NOP30,
                OpCode.OP_NOP31,
                OpCode.OP_NOP32,
                OpCode.OP_NOP33,
                OpCode.OP_NOP34,
                OpCode.OP_NOP35,
                OpCode.OP_NOP36,
                OpCode.OP_NOP37,
                OpCode.OP_NOP38,
                OpCode.OP_NOP39,
                OpCode.OP_NOP40,
                OpCode.OP_NOP41,
                OpCode.OP_NOP42,
                OpCode.OP_NOP43,
                OpCode.OP_NOP44,
                OpCode.OP_NOP45,
                OpCode.OP_NOP46,
                OpCode.OP_NOP47,
                OpCode.OP_NOP48,
                OpCode.OP_NOP49,
                OpCode.OP_NOP50,
                OpCode.OP_NOP51,
                OpCode.OP_NOP52,
                OpCode.OP_NOP53,
                OpCode.OP_NOP54,
                OpCode.OP_NOP55,
                OpCode.OP_NOP56,
                OpCode.OP_NOP57,
                OpCode.OP_NOP58,
                OpCode.OP_NOP59,
                OpCode.OP_NOP60,
                OpCode.OP_NOP61,
                OpCode.OP_NOP62,
                OpCode.OP_NOP63,
                OpCode.OP_NOP64,
                OpCode.OP_NOP65,
                OpCode.OP_NOP66,
                OpCode.OP_NOP67,
                OpCode.OP_NOP68,
                OpCode.OP_NOP69,
                OpCode.OP_NOP70,
                OpCode.OP_NOP71,
                OpCode.OP_NOP72,
                OpCode.OP_NOP73,
                OpCode.OP_NOP77,
            ]:
                pass

            elif current_opcode in [OpCode.OP_IF, OpCode.OP_NOTIF]:
                f = False
                if is_script_executing:
                    if len(self.stack) < 1:
                        _m = 'OP_IF and OP_NOTIF require at least one item on the stack when they are used!'
                        self.script_evaluation_error(_m)
                    octets = self.stacktop(-1)
                    f = self.cast_to_bool(octets)
                    if current_opcode == OpCode.OP_NOTIF:
                        f = not f
                    self.stack.pop()
                self.if_stack.append(self.encode_bool(f))

            elif current_opcode == OpCode.OP_ELSE:
                if len(self.if_stack) == 0:
                    self.script_evaluation_error('OP_ELSE requires a preceeding OP_IF.')
                f = not self.cast_to_bool(self.if_stack[-1])
                self.if_stack[-1] = self.encode_bool(f)

            elif current_opcode == OpCode.OP_ENDIF:
                if len(self.if_stack) == 0:
                    self.script_evaluation_error('OP_ENDIF requires a preceeding OP_IF.')
                self.if_stack.pop()

            elif current_opcode == OpCode.OP_VERIFY:
                if len(self.stack) < 1:
                    self.script_evaluation_error('OP_VERIFY requires at least one item to be on the stack.')
                f = self.cast_to_bool(self.stacktop(-1))
                if f:
                    self.stack.pop()
                else:
                    self.script_evaluation_error('OP_VERIFY requires the top stack value to be truthy.')

            elif current_opcode == OpCode.OP_RETURN:
                if self.context == 'UnlockingScript':
                    self.program_counter = len(self.unlocking_script.chunks)
                else:
                    self.program_counter = len(self.locking_script.chunks)
                self.if_stack = []

            elif current_opcode == OpCode.OP_TOALTSTACK:
                if len(self.stack) < 1:
                    self.script_evaluation_error('OP_TOALTSTACK requires at oeast one item to be on the stack.')
                self.alt_stack.append(self.stack.pop())

            elif current_opcode == OpCode.OP_FROMALTSTACK:
                if len(self.alt_stack) < 1:
                    self.script_evaluation_error('OP_FROMALTSTACK requires at least one item to be on the stack.')
                self.stack.append(self.alt_stack.pop())

            elif current_opcode == OpCode.OP_2DROP:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_2DROP requires at least two items to be on the stack.')
                self.stack.pop()
                self.stack.pop()

            elif current_opcode == OpCode.OP_2DUP:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_2DUP requires at least two items to be on the stack.')
                x1 = self.stacktop(-2)
                x2 = self.stacktop(-1)
                self.stack.append(x1)
                self.stack.append(x2)

            elif current_opcode == OpCode.OP_3DUP:
                if len(self.stack) < 3:
                    self.script_evaluation_error('OP_3DUP requires at least three items to be on the stack.')
                x1 = self.stacktop(-3)
                x2 = self.stacktop(-2)
                x3 = self.stacktop(-1)
                self.stack.append(x1)
                self.stack.append(x2)
                self.stack.append(x3)

            elif current_opcode == OpCode.OP_2OVER:
                if len(self.stack) < 4:
                    self.script_evaluation_error('OP_2OVER requires at least four items to be on the stack.')
                x1 = self.stacktop(-4)
                x2 = self.stacktop(-3)
                self.stack.append(x1)
                self.stack.append(x2)

            elif current_opcode == OpCode.OP_2ROT:
                if len(self.stack) < 6:
                    self.script_evaluation_error('OP_2ROT requires at least six items to be on the stack.')
                x1 = self.stack.pop(-6)
                x2 = self.stack.pop(-5)
                self.stack.append(x1)
                self.stack.append(x2)

            elif current_opcode == OpCode.OP_2SWAP:
                if len(self.stack) < 4:
                    self.script_evaluation_error('OP_2SWAP requires at least four items to be on the stack.')
                x1 = self.stack.pop(-4)
                x2 = self.stack.pop(-3)
                self.stack.append(x1)
                self.stack.append(x2)

            elif current_opcode == OpCode.OP_IFDUP:
                if len(self.stack) < 1:
                    self.script_evaluation_error('OP_IFDUP requires at least one item to be on the stack.')
                octets = self.stacktop(-1)
                f = self.cast_to_bool(octets)
                if f:
                    self.stack.append(octets)

            elif current_opcode == OpCode.OP_DEPTH:
                self.stack.append(self.minimally_encode(len(self.stack)))

            elif current_opcode == OpCode.OP_DROP:
                if len(self.stack) < 1:
                    self.script_evaluation_error('OP_DROP requires at least one item to be on the stack.')
                self.stack.pop()

            elif current_opcode == OpCode.OP_DUP:
                if len(self.stack) < 1:
                    self.script_evaluation_error('OP_DUP requires at least one item to be on the stack.')
                self.stack.append(self.stacktop(-1))

            elif current_opcode == OpCode.OP_NIP:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_NIP requires at least two items to be on the stack.')
                self.stack.pop(-2)

            elif current_opcode == OpCode.OP_OVER:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_OVER requires at least two items to be on the stack.')
                self.stack.append(self.stacktop(-2))

            elif current_opcode in [OpCode.OP_PICK, OpCode.OP_ROLL]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                if len(self.stack) < 2:
                    self.script_evaluation_error(f'{_codename} requires at least two items to be on the stack.')
                n = self.bin2num(self.stacktop(-1))
                self.stack.pop()
                if n < 0 or n >= len(self.stack):
                    _m = (f'{_codename} requires the top stack element to be 0 or '
                          'a positive number less than the current size of the stack.')
                    self.script_evaluation_error(_m)
                octets = self.stacktop(-n - 1)
                if current_opcode == OpCode.OP_ROLL:
                    octets = self.stack.pop(len(self.stack) - n - 1)
                self.stack.append(octets)

            elif current_opcode == OpCode.OP_ROT:
                if len(self.stack) < 3:
                    self.script_evaluation_error('OP_ROT requires at least three items to be on the stack.')
                x1 = self.stacktop(-3)
                x2 = self.stacktop(-2)
                x3 = self.stacktop(-1)
                self.stack[-3] = x2
                self.stack[-2] = x3
                self.stack[-1] = x1

            elif current_opcode == OpCode.OP_SWAP:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_SWAP requires at least two items to be on the stack.')
                x1 = self.stacktop(-2)
                x2 = self.stacktop(-1)
                self.stack[-2] = x2
                self.stack[-1] = x1

            elif current_opcode == OpCode.OP_TUCK:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_TUCK requires at least two items to be on the stack.')
                x1 = self.stack.pop(-2)
                x2 = self.stack.pop(-1)
                self.stack.append(x2)
                self.stack.append(x1)
                self.stack.append(x2)

            elif current_opcode == OpCode.OP_SIZE:
                if len(self.stack) < 1:
                    self.script_evaluation_error('OP_SIZE requires at least one item to be on the stack.')
                n = len(self.stacktop(-1))
                self.stack.append(self.minimally_encode(n))

            elif current_opcode in [OpCode.OP_AND, OpCode.OP_OR, OpCode.OP_XOR]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                if len(self.stack) < 2:
                    self.script_evaluation_error(f'{_codename} requires at least one item to be on the stack.')
                x1 = self.stack.pop(-2)
                x2 = self.stack.pop(-1)
                if len(x1) != len(x2):
                    self.script_evaluation_error(f'{_codename} requires the top two stack items to be the same size.')
                if current_opcode == OpCode.OP_AND:
                    sig = bytes([a & b for a, b in zip(x1, x2)])
                elif current_opcode == OpCode.OP_OR:
                    sig = bytes([a | b for a, b in zip(x1, x2)])
                else:
                    sig = bytes([a ^ b for a, b in zip(x1, x2)])
                self.stack.append(sig)

            elif current_opcode == OpCode.OP_INVERT:
                if len(self.stack) < 1:
                    self.script_evaluation_error('OP_INVERT requires at least one item to be on the stack.')
                x = self.stack.pop()
                x = bytes([~b for b in x])
                self.stack.append(x)

            elif current_opcode in [OpCode.OP_LSHIFT, OpCode.OP_RSHIFT]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                if len(self.stack) < 2:
                    self.script_evaluation_error(f'{_codename} requires at least two items to be on the stack.')
                n = self.bin2num(self.stacktop(-1))
                if n < 0:
                    self.script_evaluation_error(f'{_codename} requires the top stack item to be non-negative.')
                x = self.stack.pop(-2)
                if current_opcode == OpCode.OP_LSHIFT:
                    x = x[n:] + b'\x00' * n
                else:
                    x = b'\x00' * n + x[:-n]
                self.stack.append(x)

            elif current_opcode in [OpCode.OP_EQUAL, OpCode.OP_EQUALVERIFY]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                if len(self.stack) < 2:
                    self.script_evaluation_error(f'{_codename} requires at least two items to be on the stack.')
                x1 = self.stack.pop(-2)
                x2 = self.stack.pop(-1)
                f = x1 == x2
                self.stack.append(self.encode_bool(f))
                if current_opcode == OpCode.OP_EQUALVERIFY:
                    if f:
                        self.stack.pop()
                    else:
                        self.script_evaluation_error('OP_EQUALVERIFY requires the top two stack items to be equal.')

            elif current_opcode in [
                OpCode.OP_1ADD,
                OpCode.OP_1SUB,
                OpCode.OP_NEGATE,
                OpCode.OP_ABS,
                OpCode.OP_NOT,
                OpCode.OP_0NOTEQUAL,
            ]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                if len(self.stack) < 1:
                    self.script_evaluation_error(f'{_codename} requires at least one items to be on the stack.')
                x = self.bin2num(self.stack.pop())
                if current_opcode == OpCode.OP_1ADD:
                    x += 1
                elif current_opcode == OpCode.OP_1SUB:
                    x -= 1
                elif current_opcode == OpCode.OP_NEGATE:
                    x = -x
                elif current_opcode == OpCode.OP_ABS:
                    x = abs(x)
                elif current_opcode == OpCode.OP_NOT:
                    x = 1 if x == 0 else 0
                else:
                    x = 1 if x != 0 else 0
                self.stack.append(self.minimally_encode(x))

            elif current_opcode in [
                OpCode.OP_ADD,
                OpCode.OP_SUB,
                OpCode.OP_MUL,
                OpCode.OP_MOD,
                OpCode.OP_DIV,
                OpCode.OP_BOOLAND,
                OpCode.OP_BOOLOR,
                OpCode.OP_NUMEQUAL,
                OpCode.OP_NUMEQUALVERIFY,
                OpCode.OP_NUMNOTEQUAL,
                OpCode.OP_LESSTHAN,
                OpCode.OP_GREATERTHAN,
                OpCode.OP_LESSTHANOREQUAL,
                OpCode.OP_GREATERTHANOREQUAL,
                OpCode.OP_MIN,
                OpCode.OP_MAX,
            ]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                if len(self.stack) < 2:
                    self.script_evaluation_error(f'{_codename} requires at least two items to be on the stack.')
                x1 = self.bin2num(self.stack.pop(-2))
                x2 = self.bin2num(self.stack.pop())
                if current_opcode == OpCode.OP_ADD:
                    x = x1 + x2
                elif current_opcode == OpCode.OP_SUB:
                    x = x1 - x2
                elif current_opcode == OpCode.OP_MUL:
                    x = x1 * x2
                elif current_opcode == OpCode.OP_DIV:
                    if x2 == 0:
                        self.script_evaluation_error('OP_DIV cannot divide by zero!')
                    x = x1 // x2
                elif current_opcode == OpCode.OP_MOD:
                    if x2 == 0:
                        self.script_evaluation_error('OP_MOD cannot divide by zero!')
                    x = x1 % x2
                elif current_opcode == OpCode.OP_BOOLAND:
                    x = 1 if x1 != 0 and x2 != 0 else 0
                elif current_opcode == OpCode.OP_BOOLOR:
                    x = 1 if x1 != 0 or x2 != 0 else 0
                elif current_opcode == OpCode.OP_NUMEQUAL:
                    x = 1 if x1 == x2 else 0
                elif current_opcode == OpCode.OP_NUMEQUALVERIFY:
                    x = 1 if x1 == x2 else 0
                elif current_opcode == OpCode.OP_NUMNOTEQUAL:
                    x = 1 if x1 != x2 else 0
                elif current_opcode == OpCode.OP_LESSTHAN:
                    x = 1 if x1 < x2 else 0
                elif current_opcode == OpCode.OP_GREATERTHAN:
                    x = 1 if x1 > x2 else 0
                elif current_opcode == OpCode.OP_LESSTHANOREQUAL:
                    x = 1 if x1 <= x2 else 0
                elif current_opcode == OpCode.OP_GREATERTHANOREQUAL:
                    x = 1 if x1 >= x2 else 0
                elif current_opcode == OpCode.OP_MIN:
                    x = min(x1, x2)
                else:
                    x = max(x1, x2)
                self.stack.append(self.minimally_encode(x))

                if current_opcode == OpCode.OP_NUMEQUALVERIFY:
                    if self.cast_to_bool(self.stacktop(-1)):
                        self.stack.pop()
                    else:
                        self.script_evaluation_error('OP_NUMEQUALVERIFY requires the top stack item to be truthy.')

            elif current_opcode == OpCode.OP_WITHIN:
                if len(self.stack) < 3:
                    self.script_evaluation_error('OP_WITHIN requires at least three items to be on the stack.')
                x1 = self.bin2num(self.stack.pop(-3))
                x2 = self.bin2num(self.stack.pop(-2))
                x3 = self.bin2num(self.stack.pop())
                f = x2 <= x1 < x3
                self.stack.append(self.encode_bool(f))

            elif current_opcode in [
                OpCode.OP_RIPEMD160,
                OpCode.OP_SHA1,
                OpCode.OP_SHA256,
                OpCode.OP_HASH160,
                OpCode.OP_HASH256,
            ]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                if len(self.stack) < 1:
                    self.script_evaluation_error(f'{_codename} requires at least one item to be on the stack.')
                sig = self.stack.pop()
                if current_opcode == OpCode.OP_RIPEMD160:
                    sig = ripemd160(sig)
                elif current_opcode == OpCode.OP_SHA1:
                    sig = sha1(sig)
                elif current_opcode == OpCode.OP_SHA256:
                    sig = sha256(sig)
                elif current_opcode == OpCode.OP_HASH160:
                    sig = hash160(sig)
                else:
                    sig = hash256(sig)
                self.stack.append(sig)

            elif current_opcode == OpCode.OP_CODESEPARATOR:
                self.last_code_separator = self.program_counter

            elif current_opcode in [OpCode.OP_CHECKSIG, OpCode.OP_CHECKSIGVERIFY]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                if len(self.stack) < 2:
                    self.script_evaluation_error(f'{_codename} requires at least two items to be on the stack.')
                sig = self.stack.pop(-2)
                pub_key = self.stack.pop()
                if not self.check_signature_encoding(sig) or not self.check_public_key_encoding(pub_key):
                    _m = f'{_codename} requires correct encoding for the public key and signature.'
                    self.script_evaluation_error(_m)

                # Subset of script starting at the most recent code separator
                if self.context == 'UnlockingScript':
                    sub_script = Script.from_chunks(self.unlocking_script.chunks[self.last_code_separator:])
                else:
                    sub_script = Script.from_chunks(self.locking_script.chunks[self.last_code_separator:])

                # Drop the signature, since there's no way for a signature to sign itself
                sub_script = Script.find_and_delete(sub_script, Script.write_bin(sig))

                # TODO
                f = self.verify_signature(sig, pub_key, sub_script)

                if not f and len(sig) > 0:
                    self.script_evaluation_error(f'{_codename} failed to verify the signature, '
                                                 'and requires an empty signature when verification fails.')
                self.stack.append(self.encode_bool(f))

                if current_opcode == OpCode.OP_CHECKSIGVERIFY:
                    if f:
                        self.stack.pop()
                    else:
                        self.script_evaluation_error('OP_CHECKSIGVERIFY requires that a valid signature is provided.')

            elif current_opcode in [OpCode.OP_CHECKMULTISIG, OpCode.OP_CHECKMULTISIGVERIFY]:
                _codename = OPCODE_VALUE_NAME_DICT[current_opcode]
                i = 1
                if len(self.stack) < i:
                    self.script_evaluation_error(f'{_codename} requires at least 1 item to be on the stack.')

                keys_count = self.bin2num(self.stacktop(-i))
                if keys_count < 0 or keys_count > MAX_MULTISIG_KEY_COUNT:
                    _m = f'${_codename} requires a key count between 0 and {MAX_MULTISIG_KEY_COUNT}.'
                    self.script_evaluation_error(_m)
                i += 1
                i_key = i
                i += keys_count

                # ikey2 is the position of last non-signature item in the stack. Top stack item = 1.
                # With SCRIPT_VERIFY_NULLFAIL, this is used for cleanup if operation fails.
                i_key2 = keys_count + 2
                
                if len(self.stack) < i:
                    _m = f'{_codename} requires the number of stack items not to be less than the number of keys used.'
                    self.script_evaluation_error(_m)

                sigs_count = self.bin2num(self.stacktop(-i))
                if sigs_count < 0 or sigs_count > keys_count:
                    _m = f'{_codename} requires the number of signatures to be no greater than the number of keys.'
                    self.script_evaluation_error(_m)
                i += 1
                i_sig = i
                i += sigs_count
                if len(self.stack) < i:
                    _m = (f'{_codename} requires the number of stack items '
                          'not to be less than the number of signatures provided.')
                    self.script_evaluation_error(_m)

                # Subset of script starting at the most recent code separator
                if self.context == 'UnlockingScript':
                    sub_script = Script.from_chunks(self.unlocking_script.chunks[self.last_code_separator:])
                else:
                    sub_script = Script.from_chunks(self.locking_script.chunks[self.last_code_separator:])

                # Drop the signatures, since there's no way for a signature to sign itself
                for j in range(sigs_count):
                    sub_script = Script.find_and_delete(sub_script, Script.write_bin(self.stacktop(-i_sig - j)))

                f = True
                while f and sigs_count > 0:
                    buf_sig = self.stacktop(-i_sig)
                    buf_pub_key = self.stacktop(-i_key)

                    if not self.check_signature_encoding(buf_sig) or not self.check_public_key_encoding(buf_pub_key):
                        _m = f'{_codename} requires correct encoding for the public key and signature.'
                        self.script_evaluation_error(_m)

                    # TODO
                    f_verify = self.verify_signature(buf_sig, buf_pub_key, sub_script)

                    if f_verify:
                        i_sig += 1
                        sigs_count -= 1
                    i_key += 1
                    sigs_count -= 1

                    # If there are more signatures left than keys left, then too many signatures have failed
                    if sigs_count > keys_count:
                        f = False

                # Clean up stack of actual arguments
                while i > 1:
                    if not f and not i_key2 and ():
                        _m = (f'{_codename} failed to verify a signature, '
                              'and requires an empty signature when verification fails.')
                        self.script_evaluation_error(_m)

                    if i_key2 > 0:
                        i_key2 -= 1

                    self.stack.pop()
                    i -= 1

                # A bug causes CHECKMULTISIG to consume one extra argument whose contents were not checked in any way.
                #
                # Unfortunately this is a potential source of mutability,
                # so optionally verify it is exactly equal to zero prior
                # to removing it from the stack.
                if len(self.stack) < 1:
                    self.script_evaluation_error(f'{_codename} requires an extra item to be on the stack.')
                if len(self.stacktop(-1)) > 0:
                    self.script_evaluation_error(f'{_codename} requires the extra stack item to be empty.')
                self.stack.pop()

                self.stack.append(self.encode_bool(f))

                if current_opcode == OpCode.OP_CHECKMULTISIGVERIFY:
                    if f:
                        self.stack.pop()
                    else:
                        _m = 'OP_CHECKMULTISIGVERIFY requires a sufficient number of valid signatures are provided.'
                        self.script_evaluation_error(_m)

            elif current_opcode == OpCode.OP_CAT:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_CAT requires at least two items to be on the stack.')
                x1 = self.stack.pop(-2)
                x2 = self.stack.pop()
                if len(x1) + len(x2) > MAX_SCRIPT_ELEMENT_SIZE:
                    self.script_evaluation_error("It's not currently possible to push data "
                                                 f"larger than {MAX_SCRIPT_ELEMENT_SIZE} bytes.")
                self.stack.append(x1 + x2)

            elif current_opcode == OpCode.OP_SPLIT:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_SPLIT requires at least two items to be on the stack.')
                x1 = self.stack.pop(-2)
                #  Make sure the split point is appropriate.
                n = self.bin2num(self.stack.pop())
                if n < 0 or n > len(x1):
                    self.script_evaluation_error("OP_SPLIT requires the first stack item to be a non-negative number "
                                                 "less than or equal to the size of the second-from-top stack item.")
                self.stack.append(x1[:n])
                self.stack.append(x1[n:])

            elif current_opcode == OpCode.OP_NUM2BIN:
                if len(self.stack) < 2:
                    self.script_evaluation_error('OP_NUM2BIN requires at least two items to be on the stack.')
                size = self.bin2num(self.stack.pop())
                if size > MAX_SCRIPT_ELEMENT_SIZE:
                    self.script_evaluation_error("It's not currently possible to push data "
                                                 f"larger than {MAX_SCRIPT_ELEMENT_SIZE} bytes.")
                n = self.bin2num(self.stack.pop())
                x = bytearray(self.minimally_encode(n))

                # Try to see if we can fit that number in the number of byte requested.
                if len(x) > size:
                    _m = ('OP_NUM2BIN requires that the size expressed in the top stack item '
                          'is large enough to hold the value expressed in the second-from-top stack item.')
                    self.script_evaluation_error(_m)

                msb = b'\x00'
                if len(x) > 0:
                    msb = x[-1] & 0x80
                    x[-1] &= 0x7f
                octets = x + b'\x00' * (size - len(x))
                octets[-1] |= msb

                self.stack.append(octets)

            elif current_opcode == OpCode.OP_BIN2NUM:
                if len(self.stack) < 1:
                    self.script_evaluation_error('OP_BIN2NUM requires at least one item to be on the stack.')
                x = self.stack.pop()
                self.stack.append(self.minimally_encode(self.bin2num(x)))

            else:
                self.script_evaluation_error('Invalid opcode!')

        # Finally, increment the program counter
        self.program_counter += 1

    def validate(self) -> bool:
        """
        Validates the spend action by interpreting the locking and unlocking scripts.
        Returns true if the scripts are valid and the spend is legitimate, otherwise false.
        """
        if REQUIRE_PUSH_ONLY_UNLOCKING_SCRIPTS and not self.unlocking_script.is_push_only():
            self.script_evaluation_error('Unlocking scripts can only contain push operations, and no other opcodes.')

        while True:
            self.step()
            if self.context == 'LockingScript' and self.program_counter >= len(self.locking_script.chunks):
                break

        if len(self.if_stack) > 0:
            self.script_evaluation_error('Every OP_IF must be terminated prior to the end of the script.')

        if REQUIRE_CLEAN_STACK:
            if len(self.stack) != 1:
                self.script_evaluation_error(
                    'The clean stack rule requires exactly one item to be on the stack after script execution.')

        if not self.cast_to_bool(self.stacktop(-1)):
            self.script_evaluation_error('The top stack element must be truthy after script evaluation.')

        return True

    def stacktop(self, i: int) -> bytes:
        return self.stack[len(self.stack) + i]

    def script_evaluation_error(self, message: str) -> None:
        raise Exception(f"Script evaluation error: {message}\n\n"
                        f"Source TXID: {self.source_txid}\n"
                        f"Source output index: {self.source_output_index}\n"
                        f"Context: {self.context}\n"
                        f"Program counter: {self.program_counter}\n"
                        f"Stack size: {len(self.stack)}\n"
                        f"Alt stack size: {len(self.alt_stack)}")

    @staticmethod
    def cast_to_bool(val: bytes) -> bool:
        for i in range(len(val)):
            if val[i] != 0:
                # can be negative zero
                if i == len(val) - 1 and val[i] == 0x80:
                    return False
                return True
        return False

    @classmethod
    def is_opcode_disabled(cls, opcode: bytes) -> bool:
        return (opcode == OpCode.OP_2MUL
                or opcode == OpCode.OP_2DIV
                or opcode == OpCode.OP_VERIF
                or opcode == OpCode.OP_VERNOTIF
                or opcode == OpCode.OP_VER)

    @classmethod
    def is_chunk_minimal(cls, chunk: ScriptChunk) -> bool:
        data = chunk.data
        op = chunk.op
        if data is None:
            return True
        if len(data) == 0:
            return op == OpCode.OP_0
        if len(data) == 1 and 1 <= data[0] <= 16:
            return op == OpCode.OP_1 + (int.from_bytes(data, 'big') - 1).to_bytes(1, 'big')
        if len(data) == 1 and data[0] == 0x81:
            return op == OpCode.OP_1NEGATE
        if len(data) <= 75:
            return op == len(data).to_bytes(1, 'big')
        if len(data) <= 255:
            return op == OpCode.OP_PUSHDATA1
        if len(data) <= 65535:
            return op == OpCode.OP_PUSHDATA2
        return op == OpCode.OP_PUSHDATA4

    @classmethod
    def minimally_encode(cls, num: int) -> bytes:
        if num == 0:
            return b''
        negative: bool = num < 0
        octets: bytearray = bytearray(unsigned_to_bytes(-num if negative else num, 'little'))
        if octets[-1] & 0x80:
            octets += b'\x00'
        if negative:
            octets[-1] |= 0x80
        return octets

    @classmethod
    def bin2num(cls, octets: bytes) -> int:
        if len(octets) == 0:
            return 0
        negative = octets[-1] & 0x80
        octets = bytearray(octets)
        octets[-1] &= 0x7f
        n = int.from_bytes(octets, 'little')
        return -n if negative else n

    def check_signature_encoding(self, octets: bytes) -> bool:
        # Empty signature. Not strictly DER encoded, but allowed to provide a
        # compact way to provide an invalid signature for use with CHECK(MULTI)SIG
        if octets == b'':
            return True
        sig, sighash = octets[:-1], octets[-1]

        if not SIGHASH.validate(sighash):
            self.script_evaluation_error('Invalid SIGHASH flag')

        with suppress(Exception):
            _, s = deserialize_ecdsa_der(sig)
            if REQUIRE_LOW_S_SIGNATURES and s > curve.n // 2:
                self.script_evaluation_error('The signature must have a low S value.')
            return True
        self.script_evaluation_error('The signature format is invalid.')

    @classmethod
    def check_public_key_encoding(cls, octets: bytes) -> bool:
        with suppress(Exception):
            PublicKey(octets)
            return True
        return False

    def verify_signature(self, sig: bytes, pub_key: bytes, sub_script: Script) -> bool:
        if sig == b'':
            return False

        current_input = TransactionInput(
            source_txid=self.source_txid,
            source_output_index=self.source_output_index,
            unlocking_script=self.unlocking_script,
            sequence=self.input_sequence,
            sighash=SIGHASH(sig[-1]),
        )
        current_input.locking_script = sub_script
        current_input.satoshis = self.source_satoshis

        inputs = self.other_inputs[:]
        inputs.insert(self.input_index, current_input)

        preimage = tx_preimage(self.input_index, inputs, self.outputs, self.transaction_version, self.lock_time)
        return PublicKey(pub_key).verify(sig[:-1], preimage)

    @classmethod
    def encode_bool(cls, f: bool) -> bytes:
        return b'\x01' if f else b''



from contextlib import suppress
from typing import Optional, Union

from .script.script import Script
from .utils import Reader


class TransactionOutput:

    def __init__(
            self,
            locking_script: Script,
            satoshis: int = None,
            change: bool = False,
    ):
        self.satoshis = satoshis
        self.locking_script = locking_script
        self.change = change

    def serialize(self) -> bytes:
        return b"".join(
            [
                self.satoshis.to_bytes(8, "little"),
                self.locking_script.byte_length_varint(),
                self.locking_script.serialize(),
            ]
        )

    def __str__(self) -> str:  # pragma: no cover
        return (
            f"<TxOutput value={self.satoshis} locking_script={self.locking_script.hex()}>"
        )

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()

    @classmethod
    def from_hex(cls, stream: Union[str, bytes, Reader]) -> Optional["TransactionOutput"]:
        with suppress(Exception):
            stream = (
                stream
                if isinstance(stream, Reader)
                else Reader(
                    stream if isinstance(stream, bytes) else bytes.fromhex(stream)
                )
            )
            satoshis = stream.read_int(8)
            assert satoshis is not None
            script_length = stream.read_var_int_num()
            assert script_length is not None
            locking_script_bytes = stream.read_bytes(script_length)
            return TransactionOutput(locking_script=Script(locking_script_bytes), satoshis=satoshis)
        return None



from base64 import b64encode

from .keys import PrivateKey, PublicKey
from .curve import curve, curve_multiply
from .utils import randbytes, Reader

class SignedMessage:
    VERSION = bytes.fromhex('42423301')

    @staticmethod
    def sign(message: bytes, signer: PrivateKey, verifier: PublicKey = None) -> bytes:
        """
        Signs a message from one party to be verified by another, or for verification by anyone, using the BRC-77 message signing protocol.
        :param message: The message to sign
        :param signer: The private key of the message signer
        :param verifier: The public key of the person who can verify the message. If not provided, anyone will be able to verify the message signature.
        :return: The message signature.
        """
        recipient_anyone = verifier is None
        if recipient_anyone:
            anyone_point = curve_multiply(1, curve.g)
            verifier = PublicKey(anyone_point)

        # key_id = randbytes(32)
        key_id = bytes.fromhex('0000000000000000000000000000000000000000000000000000000000000000')
        key_id_base64 = b64encode(key_id).decode('ascii')
        invoice_number = f'2-message signing-{key_id_base64}'
        signing_key = signer.derive_child(verifier, invoice_number)
        signature = signing_key.sign(message)
        signer_public_key = signer.public_key().serialize()
        version = SignedMessage.VERSION
        return version + signer_public_key + (b'\x00' if recipient_anyone else verifier.serialize()) + key_id + signature

    @staticmethod
    def verify(message: bytes, sig: bytes, recipient: PrivateKey = None) -> bool:
        """
        Verifies a message using the BRC-77 message signing protocol.
        :param message: The message to verify.
        :param sig: The message signature to be verified.
        :param recipient: The private key of the message verifier. This can be omitted if the message is verifiable by anyone.
        :return: True if the message is verified.
        """
        reader = Reader(sig)
        message_version = reader.read(4)
        if message_version != SignedMessage.VERSION:
            raise ValueError(f'Message version mismatch: Expected {SignedMessage.VERSION.hex()}, received {message_version.hex()}')

        signer = PublicKey(reader.read(33))
        verifier_first = reader.read(1)[0]
        
        if verifier_first == 0:
            recipient = PrivateKey(1)
        else:
            verifier_rest = reader.read(32)
            verifier_der = bytes([verifier_first]) + verifier_rest
            if recipient is None:
                raise ValueError(f'This signature can only be verified with knowledge of a specific private key. The associated public key is: {verifier_der.hex()}')

            recipient_der = recipient.public_key().serialize()
            if verifier_der != recipient_der:
                raise ValueError(f'The recipient public key is {recipient_der.hex()} but the signature requires the recipient to have public key {verifier_der.hex()}')

        key_id = b64encode(reader.read(32)).decode('ascii')
        signature_der = reader.read(len(sig) - reader.tell())
        invoice_number = f'2-message signing-{key_id}'
        signing_key = signer.derive_child(recipient, invoice_number)
        print(signing_key.serialize().hex())
        return signing_key.verify(signature_der, message)


from Cryptodome.Cipher import AES


class InvalidPadding(Exception):
    pass


def append_pkcs7_padding(message: bytes) -> bytes:
    pad = 16 - (len(message) % 16)
    return message + bytes([pad]) * pad


def strip_pkcs7_padding(message: bytes) -> bytes:
    if len(message) % 16 != 0 or len(message) == 0:
        raise InvalidPadding("invalid length")
    pad = message[-1]
    if not 1 <= pad <= 16:
        raise InvalidPadding("invalid padding byte (out of range)")
    for i in message[-pad:]:
        if i != pad:
            raise InvalidPadding("invalid padding byte (inconsistent)")
    return message[0:-pad]


def aes_encrypt_with_iv(key: bytes, iv: bytes, message: bytes) -> bytes:
    return AES.new(key, AES.MODE_CBC, iv).encrypt(append_pkcs7_padding(message))


def aes_decrypt_with_iv(key: bytes, iv: bytes, message: bytes) -> bytes:
    return strip_pkcs7_padding(AES.new(key, AES.MODE_CBC, iv).decrypt(message))



import os
from enum import Enum
from typing import Dict, List

NUMBER_BYTE_LENGTH: int = 32

TRANSACTION_SEQUENCE: int = int(os.getenv('BSV_PY_SDK_TRANSACTION_SEQUENCE') or 0xffffffff)
TRANSACTION_VERSION: int = int(os.getenv('BSV_PY_SDK_TRANSACTION_VERSION') or 1)
TRANSACTION_LOCKTIME: int = int(os.getenv('BSV_PY_SDK_TRANSACTION_LOCKTIME') or 0)
TRANSACTION_FEE_RATE: float = float(os.getenv('BSV_PY_SDK_TRANSACTION_FEE_RATE') or 0.5)  # satoshi per byte
BIP39_ENTROPY_BIT_LENGTH: int = int(os.getenv('BSV_PY_SDK_BIP39_ENTROPY_BIT_LENGTH') or 128)
BIP44_DERIVATION_PATH = os.getenv('BSV_PY_SDK_BIP44_DERIVATION_PATH') or "m/44'/236'/0'"

HTTP_REQUEST_TIMEOUT: int = int(os.getenv('BSV_PY_SDK_HTTP_REQUEST_TIMEOUT') or 30)
THREAD_POOL_MAX_EXECUTORS: int = int(os.getenv('BSV_PY_SDK_THREAD_POOL_MAX_EXECUTORS') or 10)


class Network(str, Enum):
    MAINNET = 'mainnet'
    TESTNET = 'testnet'


class SIGHASH(int, Enum):
    ALL: int = 0x01
    NONE: int = 0x02
    SINGLE: int = 0x03
    ANYONECANPAY: int = 0x80

    FORKID: int = 0x40

    ALL_FORKID = ALL | FORKID
    NONE_FORKID = NONE | FORKID
    SINGLE_FORKID = SINGLE | FORKID
    ALL_ANYONECANPAY_FORKID = ALL_FORKID | ANYONECANPAY
    NONE_ANYONECANPAY_FORKID = NONE_FORKID | ANYONECANPAY
    SINGLE_ANYONECANPAY_FORKID = SINGLE_FORKID | ANYONECANPAY

    @classmethod
    def validate(cls, sighash: int) -> bool:
        return sighash in [
            cls.ALL_FORKID,
            cls.NONE_FORKID,
            cls.SINGLE_FORKID,
            cls.ALL_ANYONECANPAY_FORKID,
            cls.NONE_ANYONECANPAY_FORKID,
            cls.SINGLE_ANYONECANPAY_FORKID,
        ]


#
# P2PKH address
#
ADDRESS_MAINNET_PREFIX: bytes = b'\x00'
ADDRESS_TESTNET_PREFIX: bytes = b'\x6f'
NETWORK_ADDRESS_PREFIX_DICT: Dict[Network, bytes] = {
    Network.MAINNET: ADDRESS_MAINNET_PREFIX,
    Network.TESTNET: ADDRESS_TESTNET_PREFIX,
}
ADDRESS_PREFIX_NETWORK_DICT: Dict[bytes, Network] = {
    ADDRESS_MAINNET_PREFIX: Network.MAINNET,
    ADDRESS_TESTNET_PREFIX: Network.TESTNET,
}

#
# WIF
#
WIF_MAINNET_PREFIX: bytes = b'\x80'
WIF_TESTNET_PREFIX: bytes = b'\xef'
NETWORK_WIF_PREFIX_DICT: Dict[Network, bytes] = {
    Network.MAINNET: WIF_MAINNET_PREFIX,
    Network.TESTNET: WIF_TESTNET_PREFIX,
}
WIF_PREFIX_NETWORK_DICT: Dict[bytes, Network] = {
    WIF_MAINNET_PREFIX: Network.MAINNET,
    WIF_TESTNET_PREFIX: Network.TESTNET,
}

#
# public key
#
PUBLIC_KEY_COMPRESSED_EVEN_Y_PREFIX: bytes = b'\x02'
PUBLIC_KEY_COMPRESSED_ODD_Y_PREFIX: bytes = b'\x03'
PUBLIC_KEY_COMPRESSED_PREFIX_LIST: List[bytes] = [
    PUBLIC_KEY_COMPRESSED_EVEN_Y_PREFIX,
    PUBLIC_KEY_COMPRESSED_ODD_Y_PREFIX
]
PUBLIC_KEY_COMPRESSED_BYTE_LENGTH: int = 33
PUBLIC_KEY_UNCOMPRESSED_BYTE_LENGTH: int = 65
PUBLIC_KEY_BYTE_LENGTH_LIST: List[int] = [PUBLIC_KEY_COMPRESSED_BYTE_LENGTH, PUBLIC_KEY_UNCOMPRESSED_BYTE_LENGTH]
PUBLIC_KEY_HASH_BYTE_LENGTH: int = 20

#
# extended private key
#
XPRV_MAINNET_PREFIX: bytes = b'\x04\x88\xAD\xE4'
XPRV_TESTNET_PREFX: bytes = b'\x04\x35\x83\x94'
XPRV_PREFIX_LIST: List[bytes] = [XPRV_MAINNET_PREFIX, XPRV_TESTNET_PREFX]
NETWORK_XPRV_PREFIX_DICT: Dict[Network, bytes] = {
    Network.MAINNET: XPRV_MAINNET_PREFIX,
    Network.TESTNET: XPRV_TESTNET_PREFX,
}
XPRV_PREFIX_NETWORK_DICT: Dict[bytes, Network] = {
    XPRV_MAINNET_PREFIX: Network.MAINNET,
    XPRV_TESTNET_PREFX: Network.TESTNET,
}

#
# extended public key
#
XPUB_MAINNET_PREFIX: bytes = b'\x04\x88\xB2\x1E'
XPUB_TESTNET_PREFIX: bytes = b'\x04\x35\x87\xCF'
XPUB_PREFIX_LIST: List[bytes] = [XPUB_MAINNET_PREFIX, XPUB_TESTNET_PREFIX]
NETWORK_XPUB_PREFIX_DICT: Dict[Network, bytes] = {
    Network.MAINNET: XPUB_MAINNET_PREFIX,
    Network.TESTNET: XPUB_TESTNET_PREFIX,
}
XPUB_PREFIX_NETWORK_DICT: Dict[bytes, Network] = {
    XPUB_MAINNET_PREFIX: Network.MAINNET,
    XPUB_TESTNET_PREFIX: Network.TESTNET,
}

#
# extended key
#
XKEY_BYTE_LENGTH: int = 78
XKEY_PREFIX_LIST: List[bytes] = XPRV_PREFIX_LIST + XPUB_PREFIX_LIST
#
# BIP32
#
BIP32_SEED_BYTE_LENGTH: int = 64
#
# BIP39
#
BIP39_ENTROPY_BIT_LENGTH_LIST: List[int] = [128, 160, 192, 224, 256]


class OpCode(bytes, Enum):
    """
    https://wiki.bitcoinsv.io/index.php/Opcodes_used_in_Bitcoin_Script
    """
    OP_0 = b'\x00'
    OP_FALSE = b'\x00'
    OP_PUSHDATA1 = b'\x4c'
    OP_PUSHDATA2 = b'\x4d'
    OP_PUSHDATA4 = b'\x4e'
    OP_1NEGATE = b'\x4f'
    OP_TRUE = b'\x51'
    OP_1 = b'\x51'
    OP_2 = b'\x52'
    OP_3 = b'\x53'
    OP_4 = b'\x54'
    OP_5 = b'\x55'
    OP_6 = b'\x56'
    OP_7 = b'\x57'
    OP_8 = b'\x58'
    OP_9 = b'\x59'
    OP_10 = b'\x5a'
    OP_11 = b'\x5b'
    OP_12 = b'\x5c'
    OP_13 = b'\x5d'
    OP_14 = b'\x5e'
    OP_15 = b'\x5f'
    OP_16 = b'\x60'
    # flow control
    OP_NOP = b'\x61'
    OP_VER = b'\x62'
    OP_IF = b'\x63'
    OP_NOTIF = b'\x64'
    OP_VERIF = b'\x65'
    OP_VERNOTIF = b'\x66'
    OP_ELSE = b'\x67'
    OP_ENDIF = b'\x68'
    OP_VERIFY = b'\x69'
    OP_RETURN = b'\x6a'
    # stack
    OP_TOALTSTACK = b'\x6b'
    OP_FROMALTSTACK = b'\x6c'
    OP_2DROP = b'\x6d'
    OP_2DUP = b'\x6e'
    OP_3DUP = b'\x6f'
    OP_2OVER = b'\x70'
    OP_2ROT = b'\x71'
    OP_2SWAP = b'\x72'
    OP_IFDUP = b'\x73'
    OP_DEPTH = b'\x74'
    OP_DROP = b'\x75'
    OP_DUP = b'\x76'
    OP_NIP = b'\x77'
    OP_OVER = b'\x78'
    OP_PICK = b'\x79'
    OP_ROLL = b'\x7a'
    OP_ROT = b'\x7b'
    OP_SWAP = b'\x7c'
    OP_TUCK = b'\x7d'
    # data manipulation
    OP_CAT = b'\x7e'
    OP_SPLIT = b'\x7f'
    OP_NUM2BIN = b'\x80'
    OP_BIN2NUM = b'\x81'
    OP_SIZE = b'\x82'
    # bitwise logic
    OP_INVERT = b'\x83'
    OP_AND = b'\x84'
    OP_OR = b'\x85'
    OP_XOR = b'\x86'
    OP_EQUAL = b'\x87'
    OP_EQUALVERIFY = b'\x88'
    # arithmetic
    OP_1ADD = b'\x8b'
    OP_1SUB = b'\x8c'
    OP_2MUL = b'\x8d'
    OP_2DIV = b'\x8e'
    OP_NEGATE = b'\x8f'
    OP_ABS = b'\x90'
    OP_NOT = b'\x91'
    OP_0NOTEQUAL = b'\x92'
    OP_ADD = b'\x93'
    OP_SUB = b'\x94'
    OP_MUL = b'\x95'
    OP_DIV = b'\x96'
    OP_MOD = b'\x97'
    OP_LSHIFT = b'\x98'
    OP_RSHIFT = b'\x99'
    OP_BOOLAND = b'\x9a'
    OP_BOOLOR = b'\x9b'
    OP_NUMEQUAL = b'\x9c'
    OP_NUMEQUALVERIFY = b'\x9d'
    OP_NUMNOTEQUAL = b'\x9e'
    OP_LESSTHAN = b'\x9f'
    OP_GREATERTHAN = b'\xa0'
    OP_LESSTHANOREQUAL = b'\xa1'
    OP_GREATERTHANOREQUAL = b'\xa2'
    OP_MIN = b'\xa3'
    OP_MAX = b'\xa4'
    OP_WITHIN = b'\xa5'
    # cryptography
    OP_RIPEMD160 = b'\xa6'
    OP_SHA1 = b'\xa7'
    OP_SHA256 = b'\xa8'
    OP_HASH160 = b'\xa9'
    OP_HASH256 = b'\xaa'
    OP_CODESEPARATOR = b'\xab'
    OP_CHECKSIG = b'\xac'
    OP_CHECKSIGVERIFY = b'\xad'
    OP_CHECKMULTISIG = b'\xae'
    OP_CHECKMULTISIGVERIFY = b'\xaf'
    # reserved words
    OP_RESERVED = b'\x50'
    OP_RESERVED1 = b'\x89'
    OP_RESERVED2 = b'\x8a'
    OP_NOP1 = b'\xb0'
    OP_NOP2 = b'\xb1'
    OP_NOP3 = b'\xb2'
    OP_NOP4 = b'\xb3'
    OP_NOP5 = b'\xb4'
    OP_NOP6 = b'\xb5'
    OP_NOP7 = b'\xb6'
    OP_NOP8 = b'\xb7'
    OP_NOP9 = b'\xb8'
    OP_NOP10 = b'\xb9'
    OP_NOP11 = b'\xba'
    OP_NOP12 = b'\xbb'
    OP_NOP13 = b'\xbc'
    OP_NOP14 = b'\xbd'
    OP_NOP15 = b'\xbe'
    OP_NOP16 = b'\xbf'
    OP_NOP17 = b'\xc0'
    OP_NOP18 = b'\xc1'
    OP_NOP19 = b'\xc2'
    OP_NOP20 = b'\xc3'
    OP_NOP21 = b'\xc4'
    OP_NOP22 = b'\xc5'
    OP_NOP23 = b'\xc6'
    OP_NOP24 = b'\xc7'
    OP_NOP25 = b'\xc8'
    OP_NOP26 = b'\xc9'
    OP_NOP27 = b'\xca'
    OP_NOP28 = b'\xcb'
    OP_NOP29 = b'\xcc'
    OP_NOP30 = b'\xcd'
    OP_NOP31 = b'\xce'
    OP_NOP32 = b'\xcf'
    OP_NOP33 = b'\xd0'
    OP_NOP34 = b'\xd1'
    OP_NOP35 = b'\xd2'
    OP_NOP36 = b'\xd3'
    OP_NOP37 = b'\xd4'
    OP_NOP38 = b'\xd5'
    OP_NOP39 = b'\xd6'
    OP_NOP40 = b'\xd7'
    OP_NOP41 = b'\xd8'
    OP_NOP42 = b'\xd9'
    OP_NOP43 = b'\xda'
    OP_NOP44 = b'\xdb'
    OP_NOP45 = b'\xdc'
    OP_NOP46 = b'\xdd'
    OP_NOP47 = b'\xde'
    OP_NOP48 = b'\xdf'
    OP_NOP49 = b'\xe0'
    OP_NOP50 = b'\xe1'
    OP_NOP51 = b'\xe2'
    OP_NOP52 = b'\xe3'
    OP_NOP53 = b'\xe4'
    OP_NOP54 = b'\xe5'
    OP_NOP55 = b'\xe6'
    OP_NOP56 = b'\xe7'
    OP_NOP57 = b'\xe8'
    OP_NOP58 = b'\xe9'
    OP_NOP59 = b'\xea'
    OP_NOP60 = b'\xeb'
    OP_NOP61 = b'\xec'
    OP_NOP62 = b'\xed'
    OP_NOP63 = b'\xee'
    OP_NOP64 = b'\xef'
    OP_NOP65 = b'\xf0'
    OP_NOP66 = b'\xf1'
    OP_NOP67 = b'\xf2'
    OP_NOP68 = b'\xf3'
    OP_NOP69 = b'\xf4'
    OP_NOP70 = b'\xf5'
    OP_NOP71 = b'\xf6'
    OP_NOP72 = b'\xf7'
    OP_NOP73 = b'\xf8'
    OP_NOP77 = b'\xfc'
    # template matching params
    OP_SMALLDATA = b'\xf9'
    OP_SMALLINTEGER = b'\xfa'
    OP_PUBKEYS = b'\xfb'
    OP_PUBKEYHASH = b'\xfd'
    OP_PUBKEY = b'\xfe'
    # invalid opcode
    OP_INVALIDOPCODE = b'\xff'


OPCODE_VALUE_NAME_DICT: Dict[bytes, str] = {item.value: item.name for item in OpCode}
OPCODE_VALUE_NAME_DICT[b'\x00'] = 'OP_0'



from typing import Union, Optional, List

from bsv.constants import OpCode, OPCODE_VALUE_NAME_DICT
from bsv.utils import encode_pushdata, unsigned_to_varint, Reader


class ScriptChunk:
    """
    A representation of a chunk of a script, which includes an opcode.
    For push operations, the associated data to push onto the stack is also included.
    """

    def __init__(self, op: bytes, data: Optional[bytes] = None):
        self.op = op
        self.data = data

    def __str__(self):
        if self.data is not None:
            return self.data.hex()
        return OPCODE_VALUE_NAME_DICT[self.op]

    def __repr__(self):
        return self.__str__()


class Script:

    def __init__(self, script: Union[str, bytes, None] = None):
        """
        Create script from hex string or bytes
        """
        if script is None:
            self.script: bytes = b''
        elif isinstance(script, str):
            # script in hex string
            self.script: bytes = bytes.fromhex(script)
        elif isinstance(script, bytes):
            # script in bytes
            self.script: bytes = script
        else:
            raise TypeError('unsupported script type')
        # An array of script chunks that make up the script.
        self.chunks: List[ScriptChunk] = []
        self._build_chunks()

    def _build_chunks(self):
        self.chunks = []
        reader = Reader(self.script)
        while not reader.eof():
            op = reader.read_bytes(1)
            chunk = ScriptChunk(op)
            data = None
            if b'\x01' <= op <= b'\x4b':
                data = reader.read_bytes(int.from_bytes(op, 'big'))
            elif op == OpCode.OP_PUSHDATA1:
                data = reader.read_bytes(reader.read_int8())
            elif op == OpCode.OP_PUSHDATA2:
                data = reader.read_bytes(reader.read_int16_le())
            elif op == OpCode.OP_PUSHDATA4:
                data = reader.read_bytes(reader.read_int32_le())
            chunk.data = data
            self.chunks.append(chunk)

    def serialize(self) -> bytes:
        return self.script

    def hex(self) -> str:
        return self.script.hex()

    def byte_length(self) -> int:
        return len(self.script)

    size = byte_length

    def byte_length_varint(self) -> bytes:
        return unsigned_to_varint(self.byte_length())

    size_varint = byte_length_varint

    def is_push_only(self) -> bool:
        """
        Checks if the script contains only push data operations.
        :return: True if the script is push-only, otherwise false.
        """
        for chunk in self.chunks:
            if chunk.op > OpCode.OP_16:
                return False
        return True

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Script):
            return self.script == o.script
        return super().__eq__(o)

    def __str__(self) -> str:
        return self.script.hex()

    def __repr__(self) -> str:
        return self.__str__()

    @classmethod
    def from_chunks(cls, chunks: List[ScriptChunk]) -> 'Script':
        script = b''
        for chunk in chunks:
            script += encode_pushdata(chunk.data) if chunk.data is not None else chunk.op
        s = Script(script)
        s.chunks = chunks
        return s

    @classmethod
    def from_asm(cls, asm: str) -> 'Script':
        chunks: [ScriptChunk] = []
        tokens = asm.split(' ')
        i = 0
        while i < len(tokens):
            token = tokens[i]
            token = 'OP_0' if token == 'OP_FALSE' else token
            opcode: Optional[str] = None
            opcode_value: Optional[bytes] = None
            if token.startswith('OP_') and token in OPCODE_VALUE_NAME_DICT.values():
                opcode = token
                opcode_value = OpCode[opcode].value

            if token == '0':
                opcode_value = b'\x00'
                chunks.append(ScriptChunk(opcode_value))
                i += 1
            elif token == '-1':
                opcode_value = OpCode.OP_1NEGATE
                chunks.append(ScriptChunk(opcode_value))
                i += 1
            elif opcode is None:
                hex_string = tokens[i]
                if len(hex_string) % 2 != 0:
                    hex_string = '0' + hex_string
                hex_bytes = bytes.fromhex(hex_string)
                if hex_bytes.hex() != hex_string:
                    raise ValueError('invalid hex string in script')
                hex_len = len(hex_bytes)
                if 0 <= hex_len < int.from_bytes(OpCode.OP_PUSHDATA1, 'big'):
                    opcode_value = int.to_bytes(hex_len, 1, 'big')
                elif hex_len < pow(2, 8):
                    opcode_value = OpCode.OP_PUSHDATA1
                elif hex_len < pow(2, 16):
                    opcode_value = OpCode.OP_PUSHDATA2
                elif hex_len < pow(2, 32):
                    opcode_value = OpCode.OP_PUSHDATA4
                chunks.append(ScriptChunk(opcode_value, hex_bytes))
                i = i + 1
            elif opcode_value == OpCode.OP_PUSHDATA1 \
                    or opcode_value == OpCode.OP_PUSHDATA2 \
                    or opcode_value == OpCode.OP_PUSHDATA4:
                chunks.append(ScriptChunk(opcode_value, bytes.fromhex(tokens[i + 2])))
                i += 3
            else:
                chunks.append(ScriptChunk(opcode_value))
                i += 1
        return Script.from_chunks(chunks)

    def to_asm(self) -> str:
        return ' '.join(str(chunk) for chunk in self.chunks)

    @classmethod
    def find_and_delete(cls, source: 'Script', pattern: 'Script') -> 'Script':
        chunks = []
        for chunk in source.chunks:
            if Script.from_chunks([chunk]).hex() != pattern.hex():
                chunks.append(chunk)
        return Script.from_chunks(chunks)

    @classmethod
    def write_bin(cls, octets: bytes) -> 'Script':
        return Script(encode_pushdata(octets))



from .script import Script
from .type import ScriptTemplate, Unknown, P2PKH, OpReturn, P2PK, BareMultisig, to_unlock_script_template
from .spend import Spend
from .unlocking_template import UnlockingScriptTemplate



from contextlib import suppress
from io import BytesIO
from typing import Optional, Union

from .constants import SIGHASH
from .constants import (
    TRANSACTION_SEQUENCE,
)
from .script.script import Script
from .script.unlocking_template import UnlockingScriptTemplate
from .utils import Reader


class TransactionInput:

    def __init__(
            self,
            source_transaction=None,
            source_txid: Optional[str] = None,
            source_output_index: int = 0,
            unlocking_script: Optional[Script] = None,
            unlocking_script_template: UnlockingScriptTemplate = None,
            sequence: int = TRANSACTION_SEQUENCE,
            sighash: SIGHASH = SIGHASH.ALL_FORKID,
    ):
        utxo = None
        if source_transaction:
            utxo = source_transaction.outputs[source_output_index]

        self.source_txid = source_txid
        if source_transaction and not source_txid:
            self.source_txid = source_transaction.txid()

        self.source_output_index: int = source_output_index
        self.satoshis: int = utxo.satoshis if utxo else None
        self.locking_script: Script = utxo.locking_script if utxo else None
        self.source_transaction = source_transaction
        self.unlocking_script: Script = unlocking_script
        self.unlocking_script_template = unlocking_script_template
        self.sequence: int = sequence
        self.sighash: SIGHASH = sighash

    def serialize(self) -> bytes:
        stream = BytesIO()
        stream.write(bytes.fromhex(self.source_txid)[::-1])
        stream.write(self.source_output_index.to_bytes(4, "little"))
        stream.write(
            self.unlocking_script.byte_length_varint()
            if self.unlocking_script
            else b"\x00"
        )
        stream.write(
            self.unlocking_script.serialize() if self.unlocking_script else b""
        )
        stream.write(self.sequence.to_bytes(4, "little"))
        return stream.getvalue()

    def __str__(self) -> str:  # pragma: no cover
        return (f"<TransactionInput outpoint={self.source_txid}:{self.source_output_index} "
                f"value={self.satoshis} locking_script={self.locking_script}>")

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()

    @classmethod
    def from_hex(cls, stream: Union[str, bytes, Reader]) -> Optional["TransactionInput"]:
        with suppress(Exception):
            stream = (
                stream
                if isinstance(stream, Reader)
                else Reader(
                    stream if isinstance(stream, bytes) else bytes.fromhex(stream)
                )
            )
            txid = stream.read_bytes(32)[::-1]
            assert len(txid) == 32
            vout = stream.read_int(4)
            assert vout is not None
            script_length = stream.read_var_int_num()
            assert script_length is not None
            unlocking_script_bytes = stream.read_bytes(script_length)
            sequence = stream.read_int(4)
            assert sequence is not None

            return TransactionInput(
                source_txid=txid.hex(),
                source_output_index=vout,
                unlocking_script=Script(unlocking_script_bytes),
                sequence=sequence,
            )

        return None



from typing import Union, List

from .bip32 import Xprv, Xpub, step_to_index, ckd
from .bip39 import seed_from_mnemonic
from ..constants import Network, BIP44_DERIVATION_PATH


def derive_xprv_from_mnemonic(mnemonic: str,
                              lang: str = 'en',
                              passphrase: str = '',
                              prefix: str = 'mnemonic',
                              path: str = BIP44_DERIVATION_PATH,
                              network: Network = Network.MAINNET) -> Xprv:
    """
    derive the subtree root extended private key from mnemonic and path
    """
    seed = seed_from_mnemonic(mnemonic, lang, passphrase, prefix)
    master_xprv = Xprv.from_seed(seed, network)
    return ckd(master_xprv, path)


def derive_xkeys_from_xkey(xkey: Union[Xprv, Xpub],
                           index_start: Union[str, int],
                           index_end: Union[str, int],
                           change: Union[str, int] = 0) -> List[Union[Xprv, Xpub]]:
    """
    derive extended keys according to path "./change/index"
    """
    change_xkey = xkey.ckd(step_to_index(change))
    return [change_xkey.ckd(i) for i in range(step_to_index(index_start), step_to_index(index_end))]


def derive_xprvs_from_mnemonic(mnemonic: str,
                               index_start: Union[str, int],
                               index_end: Union[str, int],
                               lang: str = 'en',
                               passphrase: str = '',
                               prefix: str = 'mnemonic',
                               path: str = BIP44_DERIVATION_PATH,
                               change: Union[str, int] = 0,
                               network: Network = Network.MAINNET) -> List[Xprv]:
    xprv = derive_xprv_from_mnemonic(mnemonic, lang, passphrase, prefix, path, network)
    return derive_xkeys_from_xkey(xprv, index_start, index_end, change)



from abc import ABC, abstractmethod


class ChainTracker(ABC):
    """
    The Chain Tracker is responsible for verifying the validity of a given Merkle root
    for a specific block height within the blockchain.

    Chain Trackers ensure the integrity of the blockchain by
    validating new headers against the chain's history. They use accumulated
    proof-of-work and protocol adherence as metrics to assess the legitimacy of blocks.
    """

    @abstractmethod
    async def is_valid_root_for_height(self, root: str, height: int) -> bool:
        """
        Verify the validity of a Merkle root for a given block height.

        :param root: The Merkle root to verify.
        :param height: The block height to verify against.
        :return: A boolean indicating if the Merkle root is valid for the specified block height.
        """
        pass



import unittest
from unittest.mock import AsyncMock, MagicMock

from bsv.broadcaster import BroadcastResponse, BroadcastFailure
from bsv.broadcasters.arc import ARC, ARCConfig
from bsv.http_client import HttpClient, HttpResponse
from bsv.transaction import Transaction


class TestARCBroadcast(unittest.IsolatedAsyncioTestCase):

    def setUp(self):
        self.URL = "https://api.taal.com/arc"
        self.api_key = "apikey_85678993923y454i4jhd803wsd02"
        self.tx = Transaction(tx_data="Hello sCrypt")

        # Mocking the Transaction methods
        self.tx.hex = MagicMock(return_value="hexFormat")

    async def test_broadcast_success(self):
        mock_response = HttpResponse(
            ok=True,
            status_code=200,
            json_data={
                "data": {
                    "txid": "8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec",
                    "txStatus": "success",
                    "extraInfo": "extra",
                }
            },
        )
        mock_http_client = AsyncMock(HttpClient)
        mock_http_client.fetch = AsyncMock(return_value=mock_response)

        arc_config = ARCConfig(api_key=self.api_key, http_client=mock_http_client)
        arc = ARC(self.URL, arc_config)
        result = await arc.broadcast(self.tx)

        self.assertIsInstance(result, BroadcastResponse)
        self.assertEqual(
            result.txid,
            "8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec",
        )
        self.assertEqual(result.message, "success extra")

    async def test_broadcast_failure(self):
        mock_response = HttpResponse(
            ok=False,
            status_code=400,
            json_data={
                "data": {"status": "ERR_BAD_REQUEST", "detail": "Invalid transaction"}
            },
        )
        mock_http_client = AsyncMock(HttpClient)
        mock_http_client.fetch = AsyncMock(return_value=mock_response)

        arc_config = ARCConfig(api_key=self.api_key, http_client=mock_http_client)
        arc = ARC(self.URL, arc_config)
        result = await arc.broadcast(self.tx)

        self.assertIsInstance(result, BroadcastFailure)
        self.assertEqual(result.code, "400")
        self.assertEqual(result.description, "Invalid transaction")

    async def test_broadcast_exception(self):
        mock_http_client = AsyncMock(HttpClient)
        mock_http_client.fetch = AsyncMock(side_effect=Exception("Internal Error"))

        arc_config = ARCConfig(api_key=self.api_key, http_client=mock_http_client)
        arc = ARC(self.URL, arc_config)
        result = await arc.broadcast(self.tx)

        self.assertIsInstance(result, BroadcastFailure)
        self.assertEqual(result.code, "500")
        self.assertEqual(result.description, "Internal Error")


if __name__ == "__main__":
    unittest.main()



import pytest

from bsv.base58 import base58check_encode, base58check_decode, b58_encode, b58_decode
from bsv.base58 import to_base58check, from_base58check

BITCOIN_ADDRESS = '1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa'
PUBLIC_KEY_HASH = bytes.fromhex('62e907b15cbf27d5425399ebf6f0fb50ebb88f18')
MAIN_ADDRESS_PREFIX = b'\x00'


def test_base58():
    assert b58_encode(b'\x00') == '1'
    assert b58_encode(b'\x00\x00') == '11'
    assert b58_encode(b'hello world') == 'StV1DL6CwTryKyV'

    assert b58_decode('1') == b'\x00'
    assert b58_decode('111') == b'\x00\x00\x00'
    assert b58_decode('StV1DL6CwTryKyV') == b'hello world'


def test_base58check_encode():
    assert base58check_encode(b'hello world') == '3vQB7B6MrGQZaxCuFg4oh'
    assert base58check_encode(MAIN_ADDRESS_PREFIX + PUBLIC_KEY_HASH) == BITCOIN_ADDRESS


def test_base58check_decode():
    assert base58check_decode('3vQB7B6MrGQZaxCuFg4oh') == b'hello world'
    assert base58check_decode(BITCOIN_ADDRESS) == MAIN_ADDRESS_PREFIX + PUBLIC_KEY_HASH
    with pytest.raises(ValueError, match=r'invalid base58 encoded'):
        base58check_decode('l')
    with pytest.raises(ValueError, match=r'unmatched base58 checksum'):
        base58check_decode('L')


def test_to_base58check():
    payloads = [
        bytes.fromhex('f5f2d624cfb5c3f66d06123d0829d1c9cebf770e'),
        bytes.fromhex('27b5891b01da2db74cde1689a97a2acbe23d5fb1'),
        bytes.fromhex('1E99423A4ED27608A15A2616A2B0E9E52CED330AC530EDCC32C8FFC6A526AEDD'),
        bytes.fromhex('3aba4162c7251c891207b747840551a71939b0de081f85c4e44cf7c13e41daa6'),
        bytes.fromhex('086eaa677895f92d4a6c5ef740c168932b5e3f44')
    ]
    encoded = [
        '1PRTTaJesdNovgne6Ehcdu1fpEdX7913CK',
        '14cxpo3MBCYYWCgF74SWTdcmxipnGUsPw3',
        '5J3mBbAH58CpQ3Y5RNJpUKPE62SQ5tfcvU2JpbnkeyhfsYB1Jcn',
        '5JG9hT3beGTJuUAmCQEmNaxAuMacCTfXuw1R3FCXig23RQHMr4K',
        '1mayif3H2JDC62S4N3rLNtBNRAiUUP99k',
    ]
    prefixes = [
        b'\x00',
        b'\x00',
        b'\x80',
        b'\x80',
        b'\x00',
    ]
    for i in range(len(payloads)):
        assert to_base58check(payloads[i], prefixes[i]) == encoded[i]
        assert from_base58check(encoded[i]) == (prefixes[i], payloads[i])



import hashlib
import hmac

from Cryptodome.Hash import RIPEMD160


def sha1(payload: bytes) -> bytes:
    return hashlib.sha1(payload).digest()


def sha256(payload: bytes) -> bytes:
    return hashlib.sha256(payload).digest()


def double_sha256(payload: bytes) -> bytes:
    return sha256(sha256(payload))


def ripemd160(payload: bytes) -> bytes:
    return RIPEMD160.new(payload).digest()


def ripemd160_sha256(payload: bytes) -> bytes:
    return ripemd160(sha256(payload))


hash256 = double_sha256
hash160 = ripemd160_sha256


def hmac_sha256(key: bytes, message: bytes) -> bytes:
    return hmac.new(key, message, hashlib.sha256).digest()


def hmac_sha512(key: bytes, message: bytes) -> bytes:
    return hmac.new(key, message, hashlib.sha512).digest()



import pytest

from bsv.constants import OpCode, SIGHASH
from bsv.keys import PrivateKey
from bsv.script.spend import Spend
from bsv.script.script import Script
from bsv.script.type import P2PKH, OpReturn, P2PK, BareMultisig, RPuzzle
from bsv.transaction import Transaction, TransactionInput, TransactionOutput
from bsv.utils import address_to_public_key_hash, encode_pushdata, encode_int
from bsv.curve import curve_multiply, curve, Point


def test_script():
    locking_script = '76a9146a176cd51593e00542b8e1958b7da2be97452d0588ac'
    assert Script(locking_script) == Script(bytes.fromhex(locking_script))
    assert Script(locking_script).hex() == locking_script
    assert Script(locking_script).size_varint() == b'\x19'

    assert Script().serialize() == b''
    assert Script().hex() == ''
    assert Script().byte_length() == 0

    with pytest.raises(TypeError, match=r'unsupported script type'):
        # noinspection PyTypeChecker
        Script(1)


def test_p2pkh():
    address = '1AfxgwYJrBgriZDLryfyKuSdBsi59jeBX9'
    locking_script = '76a9146a176cd51593e00542b8e1958b7da2be97452d0588ac'
    assert P2PKH().lock(address) == Script(locking_script)
    assert P2PKH().lock(address_to_public_key_hash(address)) == Script(locking_script)

    with pytest.raises(TypeError, match=r"unsupported type to parse P2PKH locking script"):
        # noinspection PyTypeChecker
        P2PKH().lock(1)

    key_compressed = PrivateKey('L5agPjZKceSTkhqZF2dmFptT5LFrbr6ZGPvP7u4A6dvhTrr71WZ9')
    key_uncompressed = PrivateKey('5KiANv9EHEU4o9oLzZ6A7z4xJJ3uvfK2RLEubBtTz1fSwAbpJ2U')
    assert P2PKH().unlock(key_compressed).estimated_unlocking_byte_length() == 107
    assert P2PKH().unlock(key_uncompressed).estimated_unlocking_byte_length() == 139

    source_tx = Transaction(
        [],
        [
            TransactionOutput(
                locking_script=Script(locking_script),
                satoshis=1000
            )
        ]
    )
    tx = Transaction([
        TransactionInput(
            source_transaction=source_tx,
            source_output_index=0,
            unlocking_script_template=P2PKH().unlock(key_compressed)
        )
    ], 
    [
        TransactionOutput(
            locking_script=P2PKH().lock(address),
            change=True
        )
    ])
    
    tx.fee()
    tx.sign()
    
    unlocking_script = P2PKH().unlock(key_compressed).sign(tx, 0)
    assert isinstance(unlocking_script, Script)
    assert unlocking_script.byte_length() in [106, 107]
    
    spend = Spend({
        'sourceTXID': tx.inputs[0].source_txid,
        'sourceOutputIndex': tx.inputs[0].source_output_index,
        'sourceSatoshis': source_tx.outputs[0].satoshis,
        'lockingScript': source_tx.outputs[0].locking_script,
        'transactionVersion': tx.version,
        'otherInputs': [],
        'inputIndex': 0,
        'unlockingScript': tx.inputs[0].unlocking_script,
        'outputs': tx.outputs,
        'inputSequence': tx.inputs[0].sequence,
        'lockTime': tx.locktime,
    })
    assert spend.validate()


def test_op_return():
    assert OpReturn().lock(['0']) == Script('006a0130')
    assert OpReturn().lock(['0' * 0x4b]) == Script('006a' + '4b' + '30' * 0x4b)
    assert OpReturn().lock(['0' * 0x4c]) == Script('006a' + '4c4c' + '30' * 0x4c)
    assert OpReturn().lock(['0' * 0x0100]) == Script('006a' + '4d0001' + '30' * 0x0100)
    assert OpReturn().lock([b'\x31\x32', '345']) == Script('006a' + '023132' + '03333435')

    with pytest.raises(TypeError, match=r"unsupported type to parse OP_RETURN locking script"):
        # noinspection PyTypeChecker
        OpReturn().lock([1])


def test_p2pk():
    private_key = PrivateKey('L5agPjZKceSTkhqZF2dmFptT5LFrbr6ZGPvP7u4A6dvhTrr71WZ9')
    public_key = private_key.public_key()
    assert P2PK().lock(public_key.hex()) == P2PK().lock(public_key.serialize())

    with pytest.raises(TypeError, match=r"unsupported type to parse P2PK locking script"):
        # noinspection PyTypeChecker
        P2PK().lock(1)

    source_tx = Transaction(
        [],
        [
            TransactionOutput(
                locking_script=P2PK().lock(public_key.hex()),
                satoshis=1000
            )
        ]
    )
    tx = Transaction([
        TransactionInput(
            source_transaction=source_tx,
            source_output_index=0,
            unlocking_script_template=P2PK().unlock(private_key)
        )
    ], 
    [
        TransactionOutput(
            locking_script=P2PKH().lock(public_key.address()),
            change=True
        )
    ])

    tx.fee()
    tx.sign()

    unlocking_script = P2PK().unlock(private_key).sign(tx, 0)
    assert isinstance(unlocking_script, Script)
    assert unlocking_script.byte_length() in [72, 73]
    
    spend = Spend({
        'sourceTXID': tx.inputs[0].source_txid,
        'sourceOutputIndex': tx.inputs[0].source_output_index,
        'sourceSatoshis': source_tx.outputs[0].satoshis,
        'lockingScript': source_tx.outputs[0].locking_script,
        'transactionVersion': tx.version,
        'otherInputs': [],
        'inputIndex': 0,
        'unlockingScript': tx.inputs[0].unlocking_script,
        'outputs': tx.outputs,
        'inputSequence': tx.inputs[0].sequence,
        'lockTime': tx.locktime,
    })
    assert spend.validate()


def test_bare_multisig():
    privs = [PrivateKey(), PrivateKey(), PrivateKey()]
    pubs = [
        privs[0].public_key().serialize(), 
        privs[1].public_key().serialize(),
        privs[2].public_key().serialize()
    ]
    encoded_pks = b''.join([encode_pushdata(pk if isinstance(pk, bytes) else bytes.fromhex(pk)) for pk in pubs])

    expected_locking = encode_int(2) + encoded_pks + encode_int(3) + OpCode.OP_CHECKMULTISIG
    assert BareMultisig().lock(pubs, 2).serialize() == expected_locking
    
    source_tx = Transaction(
        [],
        [
            TransactionOutput(
                locking_script=BareMultisig().lock(pubs, 2),
                satoshis=1000
            )
        ]
    )
    tx = Transaction([
        TransactionInput(
            source_transaction=source_tx,
            source_output_index=0,
            unlocking_script_template=BareMultisig().unlock(privs[:2])
        )
    ], [
        TransactionOutput(
            locking_script=P2PKH().lock('1AfxgwYJrBgriZDLryfyKuSdBsi59jeBX9'),
            change=True
        )
    ])
    
    tx.fee()
    tx.sign()

    unlocking_script = BareMultisig().unlock(privs[:2]).sign(tx, 0)
    assert isinstance(unlocking_script, Script)
    assert unlocking_script.byte_length() >= 144
    
    spend = Spend({
        'sourceTXID': tx.inputs[0].source_txid,
        'sourceOutputIndex': tx.inputs[0].source_output_index,
        'sourceSatoshis': source_tx.outputs[0].satoshis,
        'lockingScript': source_tx.outputs[0].locking_script,
        'transactionVersion': tx.version,
        'otherInputs': [],
        'inputIndex': 0,
        'unlockingScript': tx.inputs[0].unlocking_script,
        'outputs': tx.outputs,
        'inputSequence': tx.inputs[0].sequence,
        'lockTime': tx.locktime,
    })
    assert spend.validate()


def test_is_push_only():
    assert Script('00').is_push_only()  # OP_0
    assert not Script('006a').is_push_only()  # OP_0 OP_RETURN
    assert Script('4c051010101010').is_push_only()

    # like bitcoind, we regard OP_RESERVED as being "push only"
    assert Script('50').is_push_only()  # OP_RESERVED


def test_to_asm():
    assert Script('000301020300').to_asm() == 'OP_0 010203 OP_0'

    asm = 'OP_DUP OP_HASH160 f4c03610e60ad15100929cc23da2f3a799af1725 OP_EQUALVERIFY OP_CHECKSIG'
    assert Script('76a914f4c03610e60ad15100929cc23da2f3a799af172588ac').to_asm() == asm


def test_from_asm():
    assert Script.from_asm('OP_0 3 010203 OP_0').to_asm() == 'OP_0 03 010203 OP_0'

    asms = [
        '',
        'OP_0 010203 OP_0',
        'OP_SHA256 8cc17e2a2b10e1da145488458a6edec4a1fdb1921c2d5ccbc96aa0ed31b4d5f8 OP_EQUALVERIFY',
    ]
    for asm in asms:
        assert Script.from_asm(asm).to_asm() == asm

    _asm_pushdata(220)
    _asm_pushdata(1024)
    _asm_pushdata(pow(2, 17))

    asms = [
        'OP_FALSE',
        'OP_0',
        '0',
    ]
    for asm in asms:
        assert Script.from_asm(asm).to_asm() == 'OP_0'

    asms = [
        'OP_1NEGATE',
        '-1',
    ]
    for asm in asms:
        assert Script.from_asm(asm).to_asm() == 'OP_1NEGATE'


def _asm_pushdata(byte_length: int):
    octets = b'\x00' * byte_length
    asm = 'OP_RETURN ' + octets.hex()
    assert Script.from_asm(asm).to_asm() == asm


def test_find_and_delete():
    source = Script.from_asm('OP_RETURN f0f0')
    assert Script.find_and_delete(source, Script.from_asm('f0f0')).to_asm() == 'OP_RETURN'
    
def test_r_puzzle():
    private_key = PrivateKey()
    public_key = private_key.public_key()
    
    k = PrivateKey().int()
    G: Point = curve.g
    r = curve_multiply(k, G).x % curve.n
    
    r_bytes = r.to_bytes(32, byteorder='big')
    if r_bytes[0] > 0x7f:
        r_bytes = b'\x00' + r_bytes
    
    source_tx = Transaction(
        [],
        [
            TransactionOutput(
                locking_script=RPuzzle().lock(r_bytes), satoshis=100
            ),
            TransactionOutput(
                locking_script=P2PKH().lock(private_key.address()), change=True
            )
        ]
    )
    
    source_tx.fee()
    source_tx.sign()
    
    tx = Transaction(
        [
            TransactionInput(
                source_transaction=source_tx,
                source_txid=source_tx.txid(),
                source_output_index=0,
                unlocking_script_template=RPuzzle().unlock(k),
            )
        ],
        [
            TransactionOutput(
                locking_script=P2PKH().lock(private_key.address()), change=True
            )
        ]
    )
    
    tx.fee()
    tx.sign()

    assert(len(tx.inputs[0].unlocking_script.serialize()) >= 106)
    
    spend = Spend({
        'sourceTXID': tx.inputs[0].source_txid,
        'sourceOutputIndex': tx.inputs[0].source_output_index,
        'sourceSatoshis': source_tx.outputs[0].satoshis,
        'lockingScript': source_tx.outputs[0].locking_script,
        'transactionVersion': tx.version,
        'otherInputs': [],
        'inputIndex': 0,
        'unlockingScript': tx.inputs[0].unlocking_script,
        'outputs': tx.outputs,
        'inputSequence': tx.inputs[0].sequence,
        'lockTime': tx.locktime,
    })
    assert spend.validate()

def test_p2pkh_sighash_acp():
    key = PrivateKey()

    source_tx = Transaction(
        [],
        [
            TransactionOutput(
                locking_script=P2PKH().lock(key.address()),
                satoshis=1000
            ),
            TransactionOutput(
                locking_script=P2PKH().lock(key.address()),
                satoshis=245
            )
        ]
    )
    tx = Transaction([
        TransactionInput(
            source_transaction=source_tx,
            source_output_index=0,
            unlocking_script_template=P2PKH().unlock(key),
            sighash=SIGHASH.ALL_ANYONECANPAY_FORKID
        )
    ], 
    [
        TransactionOutput(
            locking_script=P2PKH().lock(key.address()),
            change=True
        )
    ])
    
    tx.fee()
    tx.sign()
    
    # Add another input that shouldn't break signature.
    tx.add_input(
        TransactionInput(
            source_transaction=source_tx,
            source_output_index=1,
            unlocking_script_template=P2PKH().unlock(key)
        )
    )
    
    spend = Spend({
        'sourceTXID': tx.inputs[0].source_txid,
        'sourceOutputIndex': tx.inputs[0].source_output_index,
        'sourceSatoshis': source_tx.outputs[0].satoshis,
        'lockingScript': source_tx.outputs[0].locking_script,
        'transactionVersion': tx.version,
        'otherInputs': [tx.inputs[1]],
        'inputIndex': 0,
        'unlockingScript': tx.inputs[0].unlocking_script,
        'outputs': tx.outputs,
        'inputSequence': tx.inputs[0].sequence,
        'lockTime': tx.locktime,
    })
    assert spend.validate()


from .bip32 import Xkey, Xprv, Xpub, ckd, step_to_index, master_xprv_from_seed
from .bip39 import WordList, mnemonic_from_entropy, seed_from_mnemonic, validate_mnemonic
from .bip44 import derive_xkeys_from_xkey, derive_xprvs_from_mnemonic, derive_xprv_from_mnemonic



import json
import random
from typing import Optional, Dict, Union, TYPE_CHECKING

from ..broadcaster import BroadcastResponse, BroadcastFailure, Broadcaster
from ..http_client import HttpClient, default_http_client


if TYPE_CHECKING:
    from ..transaction import Transaction

def to_hex(bytes_data):
    return "".join(f"{x:02x}" for x in bytes_data)


def random_hex(length: int) -> str:
    return "".join(f"{random.randint(0, 255):02x}" for _ in range(length))


class ARCConfig:
    def __init__(
        self,
        api_key: Optional[str] = None,
        http_client: Optional[HttpClient] = None,
        deployment_id: Optional[str] = None,
        callback_url: Optional[str] = None,
        callback_token: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ):
        self.api_key = api_key
        self.http_client = http_client
        self.deployment_id = deployment_id
        self.callback_url = callback_url
        self.callback_token = callback_token
        self.headers = headers


def default_deployment_id() -> str:
    return f"py-sdk-{random_hex(16)}"


class ARC(Broadcaster):
    def __init__(self, url: str, config: Union[str, ARCConfig] = None):
        self.URL = url
        if isinstance(config, str):
            self.api_key = config
            self.http_client = default_http_client()
            self.deployment_id = default_deployment_id()
            self.callback_url = None
            self.callback_token = None
            self.headers = None
        else:
            config = config or ARCConfig()
            self.api_key = config.api_key
            self.http_client = config.http_client or default_http_client()
            self.deployment_id = config.deployment_id or default_deployment_id()
            self.callback_url = config.callback_url
            self.callback_token = config.callback_token
            self.headers = config.headers

    async def broadcast(
        self, tx: 'Transaction'
    ) -> Union[BroadcastResponse, BroadcastFailure]:
        request_options = {
            "method": "POST",
            "headers": self.request_headers(),
            "data": {"rawTx": tx.to_ef().hex()},
        }

        try:
            response = await self.http_client.fetch(
                f"{self.URL}/v1/tx", request_options
            )
            
            response_json = response.json()
            
            if response.ok and response.status_code >= 200 and response.status_code <= 299:
                data = response_json["data"]

                if data.get("txid"):
                    return BroadcastResponse(
                        status="success",
                        txid=data.get("txid"),
                        message=f"{data.get('txStatus', '')} {data.get('extraInfo', '')}",
                    )
                else:
                    return BroadcastFailure(
                        status="failure",
                        code=data.get("status", "ERR_UNKNOWN"),
                        description=data.get("detail", "Unknown error"),
                    )
            else:
                return BroadcastFailure(
                    status="failure",
                    code=str(response.status_code),
                    description=response_json["data"]["detail"] if "data" in response_json else "Unknown error",
                )
            
        except Exception as error:
            return BroadcastFailure(
                status="failure",
                code="500",
                description=(
                    str(error)
                    if isinstance(error, Exception)
                    else "Internal Server Error"
                ),
            )

    def request_headers(self) -> Dict[str, str]:
        headers = {
            "Content-Type": "application/json",
            "XDeployment-ID": self.deployment_id,
        }

        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"

        if self.callback_url:
            headers["X-CallbackUrl"] = self.callback_url

        if self.callback_token:
            headers["X-CallbackToken"] = self.callback_token

        if self.headers:
            headers.update(self.headers)

        return headers



from setuptools import setup, find_packages

with open('bsv/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' ').strip("'")
            break

setup(
    version=version,
    packages=find_packages(exclude=('tests',)),
)



from base64 import b64encode

from Cryptodome.Cipher import AES

from .keys import PrivateKey, PublicKey
from .utils import randbytes

class EncryptedMessage:
    VERSION = bytes.fromhex('42421033')

    @staticmethod
    def aes_gcm_encrypt(key: bytes, message: bytes) -> bytes:
        iv = randbytes(32)
        encrypted, auth_tag = AES.new(key, AES.MODE_GCM, iv).encrypt_and_digest(message)
        return iv + encrypted + auth_tag

    @staticmethod
    def aes_gcm_decrypt(key: bytes, message: bytes) -> bytes:
        iv, encrypted, auth_tag = message[:32], message[32:-16], message[-16:]
        return AES.new(key, AES.MODE_GCM, iv).decrypt_and_verify(encrypted, auth_tag)

    @classmethod
    def encrypt(cls, message: bytes, sender: PrivateKey, recipient: PublicKey) -> bytes:
        key_id = randbytes(32)
        key_id_base64 = b64encode(key_id).decode('ascii')
        invoice_number = f'2-message encryption-{key_id_base64}'
        sender_child = sender.derive_child(recipient, invoice_number)
        recipient_child = recipient.derive_child(sender, invoice_number)
        shared_secret = sender_child.derive_shared_secret(recipient_child)
        symmetric_key = shared_secret[1:]
        encrypted = cls.aes_gcm_encrypt(symmetric_key, message)
        return cls.VERSION + sender.public_key().serialize() + recipient.serialize() + key_id + encrypted

    @classmethod
    def decrypt(cls, message: bytes, recipient: PrivateKey) -> bytes:
        try:
            version = message[:4]
            sender_pubkey, recipient_pubkey = message[4:37], message[37:70]
            key_id = message[70:102]
            encrypted = message[102:]
            if version != cls.VERSION:
                raise ValueError(f'message version mismatch, expected {cls.VERSION.hex()} but got {version.hex()}')
            if recipient_pubkey != recipient.public_key().serialize():
                _expected = recipient.public_key().hex()
                _actual = recipient_pubkey.hex()
                raise ValueError(f'recipient public key mismatch, expected {_expected} but got {_actual}')
            key_id_base64 = b64encode(key_id).decode('ascii')
            invoice_number = f'2-message encryption-{key_id_base64}'
            sender = PublicKey(sender_pubkey)
            sender_child = sender.derive_child(recipient, invoice_number)
            recipient_child = recipient.derive_child(sender, invoice_number)
            shared_secret = sender_child.derive_shared_secret(recipient_child)
            symmetric_key = shared_secret[1:]
            return cls.aes_gcm_decrypt(symmetric_key, encrypted)
        except Exception as e:
            raise ValueError(f'failed to decrypt message: {e}') from e


from .broadcasters import *
from .broadcaster import *
from .chaintrackers import *
from .chaintracker import *
from .constants import *
from .curve import *
from .fee_models import *
from .fee_model import *
from .script import * 
from .hash import *
from .utils import *
from .transaction_preimage import *
from .http_client import HttpClient, default_http_client
from .keys import verify_signed_text, PublicKey, PrivateKey
from .merkle_path import MerklePath, MerkleLeaf
from .transaction import Transaction, InsufficientFunds
from .transaction_input import TransactionInput
from .transaction_output import TransactionOutput
from .encrypted_message import *
from .signed_message import *


__version__ = '0.5.2'



from abc import abstractmethod, ABCMeta
from typing import Union, List, Optional

from .script import Script
from .unlocking_template import UnlockingScriptTemplate
from ..constants import (
    OpCode,
    PUBLIC_KEY_HASH_BYTE_LENGTH,
    PUBLIC_KEY_BYTE_LENGTH_LIST,
    SIGHASH
)
from ..keys import PrivateKey
from ..utils import address_to_public_key_hash, encode_pushdata, encode_int
from ..hash import hash256


def to_unlock_script_template(sign, estimated_unlocking_byte_length):
    class_attrs = {"sign": sign, "estimated_unlocking_byte_length": estimated_unlocking_byte_length}

    dynamic_class = type("UnlockScriptTemplateImpl", (UnlockingScriptTemplate,), class_attrs)

    return dynamic_class


class ScriptTemplate(metaclass=ABCMeta):

    @abstractmethod
    def lock(self, **kwargs) -> Script:
        """
        :returns: locking script
        """
        raise NotImplementedError("ScriptTemplate.locking")

    @abstractmethod
    def unlock(self, **kwargs) -> UnlockingScriptTemplate:
        """
        :returns: sign (function), estimated_unlocking_byte_length (function)
        """
        raise NotImplementedError("ScriptTemplate.unlocking")


class Unknown(ScriptTemplate):  # pragma: no cover

    def __str__(self) -> str:
        return "<ScriptTemplate:Unknown>"

    def __repr__(self) -> str:
        return self.__str__()

    def lock(self, **kwargs) -> Script:
        raise ValueError("don't know how to lock for script of unknown type")

    def unlock(self, **kwargs):
        raise ValueError("don't know how to unlock for script of unknown type")


class P2PKH(ScriptTemplate):

    def __str__(self) -> str:  # pragma: no cover
        return "<ScriptTemplate:P2PKH>"

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()

    def lock(self, addr: Union[str, bytes]) -> Script:
        """
        from address (str) or public key hash160 (bytes)
        """
        if isinstance(addr, str):
            pkh: bytes = address_to_public_key_hash(addr)
        elif isinstance(addr, bytes):
            pkh: bytes = addr
        else:
            raise TypeError("unsupported type to parse P2PKH locking script")

        assert (
                len(pkh) == PUBLIC_KEY_HASH_BYTE_LENGTH
        ), "invalid byte length of public key hash"

        return Script(
            OpCode.OP_DUP
            + OpCode.OP_HASH160
            + encode_pushdata(pkh)
            + OpCode.OP_EQUALVERIFY
            + OpCode.OP_CHECKSIG
        )

    def unlock(self, private_key: PrivateKey):
        def sign(tx, input_index) -> Script:
            tx_input = tx.inputs[input_index]
            sighash = tx_input.sighash

            signature = private_key.sign(tx.preimage(input_index))

            public_key: bytes = private_key.public_key().serialize()
            return Script(
                encode_pushdata(signature + sighash.to_bytes(1, "little"))
                + encode_pushdata(public_key)
            )

        def estimated_unlocking_byte_length() -> int:
            return 107 if private_key.compressed else 139

        return to_unlock_script_template(sign, estimated_unlocking_byte_length)


class OpReturn(ScriptTemplate):

    def __str__(self) -> str:  # pragma: no cover
        return "<ScriptTemplate:OP_RETURN>"

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()

    def lock(self, pushdatas: List[Union[str, bytes]]) -> Script:
        script: bytes = OpCode.OP_FALSE + OpCode.OP_RETURN
        for pushdata in pushdatas:
            if isinstance(pushdata, str):
                pushdata_bytes: bytes = pushdata.encode("utf-8")
            elif isinstance(pushdata, bytes):
                pushdata_bytes: bytes = pushdata
            else:
                raise TypeError("unsupported type to parse OP_RETURN locking script")
            script += encode_pushdata(pushdata_bytes, minimal_push=False)
        return Script(script)

    def unlock(self, **kwargs):  # pragma: no cover
        raise ValueError("OP_RETURN cannot be unlocked")


class P2PK(ScriptTemplate):

    def __str__(self) -> str:  # pragma: no cover
        return "<ScriptTemplate:P2PK>"

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()

    def lock(self, public_key: Union[str, bytes]) -> Script:
        """
        from public key in format str or bytes
        """
        if isinstance(public_key, str):
            pk: bytes = bytes.fromhex(public_key)
        elif isinstance(public_key, bytes):
            pk: bytes = public_key
        else:
            raise TypeError("unsupported type to parse P2PK locking script")

        assert (
                len(pk) in PUBLIC_KEY_BYTE_LENGTH_LIST
        ), "invalid byte length of public key"

        return Script(encode_pushdata(pk) + OpCode.OP_CHECKSIG)

    def unlock(self, private_key: PrivateKey):
        def sign(tx, input_index) -> Script:
            tx_input = tx.inputs[input_index]
            sighash = tx_input.sighash

            signature = private_key.sign(tx.preimage(input_index))
            return Script(encode_pushdata(signature + sighash.to_bytes(1, "little")))

        def estimated_unlocking_byte_length() -> int:
            return 73

        return to_unlock_script_template(sign, estimated_unlocking_byte_length)


class BareMultisig(ScriptTemplate):

    def __str__(self) -> str:  # pragma: no cover
        return "<ScriptTemplate:BareMultisig>"

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()

    def lock(self, participants: List[Union[str, bytes]], threshold: int) -> Script:
        assert (
                1 <= threshold <= len(participants)
        ), "bad threshold or number of participants"

        participants_parsed = []
        for participant in participants:
            assert type(participant).__name__ in [
                "str",
                "bytes",
            ], "unsupported public key type"
            if isinstance(participant, str):
                participant = bytes.fromhex(participant)
            assert (
                    len(participant) in PUBLIC_KEY_BYTE_LENGTH_LIST
            ), "invalid byte length of public key"
            participants_parsed.append(participant)
        script: bytes = encode_int(threshold)
        for participant in participants_parsed:
            script += encode_pushdata(participant)
        return Script(script + encode_int(len(participants)) + OpCode.OP_CHECKMULTISIG)

    def unlock(self, private_keys: List[PrivateKey]):
        def sign(tx, input_index) -> Script:
            tx_input = tx.inputs[input_index]
            sighash = tx_input.sighash

            script: bytes = OpCode.OP_0 # Append 0 to satisfy SCRIPT_VERIFY_NULLDUMMY
            for private_key in private_keys:
                signature = private_key.sign(tx.preimage(input_index))
                script += encode_pushdata(signature + sighash.to_bytes(1, "little"))
            return Script(script) 

        def estimated_unlocking_byte_length() -> int:
            return 1 + 73 * len(private_keys) + 1

        return to_unlock_script_template(sign, estimated_unlocking_byte_length)
    
class RPuzzle(ScriptTemplate):
    
    def __init__(self, puzzle_type: str = 'raw'):
        """
        Constructs an R Puzzle template instance for a given puzzle type.

        :param puzzle_type: Denotes the type of puzzle to create ('raw', 'SHA1', 'SHA256', 'HASH256', 'RIPEMD160', 'HASH160')
        """
        assert(puzzle_type in ['raw', 'SHA1', 'SHA256', 'HASH256', 'RIPEMD160', 'HASH160'])
        self.type = puzzle_type

    def lock(self, value: bytes) -> Script:
        """
        Creates an R puzzle locking script for a given R value or R value hash.

        :param value: A byte array representing the R value or its hash.
        :returns: An R puzzle locking script.
        """
        chunks = [
            OpCode.OP_OVER,
            OpCode.OP_3,
            OpCode.OP_SPLIT,
            OpCode.OP_NIP,
            OpCode.OP_1,
            OpCode.OP_SPLIT,
            OpCode.OP_SWAP,
            OpCode.OP_SPLIT,
            OpCode.OP_DROP
        ]
        if self.type != 'raw':
            chunks.append(getattr(OpCode, f'OP_{self.type}'))
        chunks.append(encode_pushdata(value))
        chunks.append(OpCode.OP_EQUALVERIFY)
        chunks.append(OpCode.OP_CHECKSIG)
        return Script(b''.join(chunks))
    
    
    def unlock(self, k: int, private_key: Optional[PrivateKey] = PrivateKey(), sign_outputs: str = 'all', anyone_can_pay: bool = False):
        """
        Creates a function that generates an R puzzle unlocking script along with its signature and length estimation.

        :param k: The K-value used to unlock the R-puzzle.
        :param private_key: The private key used for signing the transaction.
        :param sign_outputs: The signature scope for outputs ('all', 'none', 'single').
        :param anyone_can_pay: Flag indicating if the signature allows for other inputs to be added later.
        :returns: An object containing the `sign` and `estimate_length` functions.
        """
        def sign(tx, input_index) -> Script:
            sighash = SIGHASH.FORKID
            if sign_outputs == 'all':
                sighash |= SIGHASH.ALL
            elif sign_outputs == 'none':
                sighash |= SIGHASH.NONE
            elif sign_outputs == 'single':
                sighash |= SIGHASH.SINGLE
            if anyone_can_pay:
                sighash |= SIGHASH.ANYONECANPAY
                
            tx.inputs[input_index].sighash = sighash

            preimage = tx.preimage(input_index)

            sig = private_key.sign(preimage, hasher=hash256, k=k) + sighash.to_bytes(1, "little")
            pubkey_for_script = private_key.public_key().serialize()

            return Script(encode_pushdata(sig) + encode_pushdata(pubkey_for_script))

        def estimated_unlocking_byte_length() -> int:
            # public key (1+33) + signature (1+73)
            # Note: We add 1 to each element's length because of the associated OP_PUSH
            return 108

        return to_unlock_script_template(sign, estimated_unlocking_byte_length)
